import pytest
from redis.asyncio import Redis

from harp.config import ConfigurationBuilder
from harp.models import Blob
from harp.services import CannotResolveTypeException
from harp_apps.storage.services.blob_storages.redis import RedisBlobStorage


async def test_basics(redis_blob_storage: RedisBlobStorage):
    _storage = redis_blob_storage
    assert _storage.client is not None
    blob = Blob(id="foo", data=b"bar", content_type="text/plain")

    assert _storage.type == "redis"
    assert await _storage.get("foo") is None
    assert await _storage.client.get("foo") is None

    assert await _storage.put(blob) == blob
    assert await _storage.get("foo") == blob
    assert await _storage.client.get("foo") == b"text/plain\nbar"

    assert await _storage.exists("foo")
    assert await _storage.delete("foo") is None
    assert not await _storage.exists("foo")
    assert await _storage.get("foo") is None
    assert await _storage.client.get("foo") is None


class SystemTestMixin:
    applications = []
    settings = {}

    @pytest.fixture
    async def system(self):
        system = await ConfigurationBuilder(
            {
                "applications": self.applications,
                **self.settings,
            },
            use_default_applications=False,
        ).abuild_system()
        try:
            yield system
        finally:
            await system.dispose()


class TestRedisBlobStorageServiceDefaults(SystemTestMixin):
    applications = ["storage"]
    settings = {
        "storage": {
            "blobs": {"type": "redis"},
        },
    }

    async def test_load_redis_blob_storage_service(self, system):
        redis = system.provider.get("storage.redis")
        assert isinstance(redis, Redis)
        assert redis.connection_pool.connection_kwargs == {
            "db": 0,
            "host": "localhost",
            "port": 6379,
        }
        assert system.provider.get("storage.blobs")


class TestRedisBlobStorageServiceCustom(SystemTestMixin):
    applications = ["storage"]
    settings = {
        "storage": {
            "blobs": {
                "type": "redis",
            },
            "redis": {
                "url": "redis://redis.example.com:1234/1",
            },
        }
    }

    async def test_load_redis_blob_storage_service(self, system):
        redis = system.provider.get("storage.redis")
        assert isinstance(redis, Redis)
        assert redis.connection_pool.connection_kwargs == {
            "db": 1,
            "host": "redis.example.com",
            "port": 1234,
        }
        assert system.provider.get("storage.blobs")


class TestRedisBlobStorageNotConfigured(SystemTestMixin):
    applications = ["storage"]
    settings = {
        "storage": {
            "blobs": {
                "type": "sql",
            },
        }
    }

    async def test_sql_does_not_load_redis_services(self, system):
        try:
            with pytest.raises(CannotResolveTypeException):
                system.provider.get("storage.redis")
        finally:
            await system.dispose()



from datetime import UTC, datetime

from harp.models import Blob, Message, Transaction
from harp.utils.guids import generate_transaction_id_ksuid
from harp_apps.storage.services.sql import SqlStorage
from harp_apps.storage.types import IBlobStorage


class StorageTestFixtureMixin:
    async def create_transaction(self, sql_storage: SqlStorage, **kwargs):
        return await sql_storage.transactions.create(
            Transaction(
                **{
                    **{
                        "id": generate_transaction_id_ksuid(),
                        "type": "http",
                        "endpoint": "/",
                        "started_at": datetime.now(UTC),
                    },
                    **kwargs,
                }
            )
        )

    async def create_blob(self, blob_storage: IBlobStorage, data, /, **kwargs):
        return await blob_storage.put(Blob.from_data(data, **kwargs))

    async def create_message(self, sql_storage: SqlStorage, **kwargs):
        return await sql_storage.messages.create(Message(**kwargs))



import hishel
import httpx
from httpx import AsyncClient

from harp.config import ConfigurationBuilder
from harp.utils.testing.config import BaseConfigurableTest
from harp_apps.http_client.settings import CacheSettings, HttpClientSettings
from harp_apps.http_client.transport import AsyncFilterableTransport


class BaseHttpClientSettingsTest(BaseConfigurableTest):
    type = HttpClientSettings

    async def create_system(self, settings: HttpClientSettings, /):
        builder = ConfigurationBuilder(
            {
                "applications": ["http_client"],
                "http_client": settings,
            },
            use_default_applications=False,
        )

        return await builder.abuild_system()


class TestHttpClientSettings(BaseHttpClientSettingsTest):
    expected_verbose = {
        "cache": {
            "controller": {
                "allow_heuristics": False,
                "allow_stale": False,
                "cacheable_methods": ["GET", "HEAD"],
                "cacheable_status_codes": [
                    200,
                    203,
                    204,
                    206,
                    300,
                    301,
                    308,
                    404,
                    405,
                    410,
                    414,
                    501,
                ],
                "type": "hishel.Controller",
            },
            "enabled": True,
            "storage": {
                "base": "hishel.AsyncBaseStorage",
                "check_ttl_every": 60.0,
                "ttl": None,
                "type": "harp_apps.http_client.contrib.hishel.storages.AsyncStorage",
            },
            "transport": {"type": "hishel.AsyncCacheTransport"},
        },
        "proxy_transport": {"type": "harp_apps.http_client.transport.AsyncFilterableTransport"},
        "timeout": 30.0,
        "transport": {"retries": 0, "type": "httpx.AsyncHTTPTransport", "verify": True},
        "type": "httpx.AsyncClient",
    }

    async def test_without_cache(self):
        settings = self.create(cache={"enabled": False})
        assert settings.cache.enabled is False

        system = await self.create_system(settings)
        assert system.config["http_client"].cache.enabled is False

        http_client = system.provider.get("http_client")
        assert http_client is system.provider.get(AsyncClient)
        assert isinstance(http_client._transport, AsyncFilterableTransport)
        assert isinstance(http_client._transport._transport, httpx.AsyncHTTPTransport)

    async def test_without_cache_with_custom_client(self):
        settings = self.create(
            transport={"type": "httpx._client.BaseClient", "arguments": {}},
            cache={"enabled": False},
        )
        assert settings.cache.enabled is False

        system = await self.create_system(settings)
        http_client = system.provider.get("http_client")

        assert isinstance(http_client._transport._transport, httpx._client.BaseClient)

    async def test_with_default_cache(self):
        settings = self.create()

        assert settings.cache.enabled is True
        assert isinstance(settings.cache, CacheSettings)

        system = await self.create_system(settings)
        http_client = system.provider.get("http_client")

        assert type(http_client._transport).__name__ == "AsyncCacheTransport"
        assert http_client._transport._controller._allow_heuristics is False
        assert http_client._transport._controller._allow_stale is False
        assert http_client._transport._controller._cacheable_methods == ["GET", "HEAD"]
        assert http_client._transport._controller._cacheable_status_codes == list(
            hishel.HEURISTICALLY_CACHEABLE_STATUS_CODES
        )

    async def test_with_custom_cache(self):
        settings = HttpClientSettings(
            cache={
                "enabled": True,
                "controller": {
                    "allow_stale": True,
                    "cacheable_methods": ["GET"],
                    "cacheable_status_codes": [200],
                },
            }
        )

        assert settings.cache.enabled is True
        assert isinstance(settings.cache, CacheSettings)

        system = await self.create_system(settings)
        http_client = system.provider.get("http_client")

        assert type(http_client._transport).__name__ == "AsyncCacheTransport"

        assert isinstance(http_client._transport._controller, hishel.Controller)
        assert http_client._transport._controller._allow_heuristics is False
        assert http_client._transport._controller._allow_stale is True
        assert http_client._transport._controller._cacheable_methods == ["GET"]
        assert http_client._transport._controller._cacheable_status_codes == [200]



from sqlalchemy import TIMESTAMP, LargeBinary, String, delete, func, select
from sqlalchemy.orm import aliased, mapped_column

from harp.models import Blob as BlobModel

from .base import Base, Repository, with_session
from .messages import Message


class Blob(Base):
    __tablename__ = "blobs"

    id = mapped_column(String(40), primary_key=True, unique=True)
    data = mapped_column(LargeBinary())
    content_type = mapped_column(String(64))
    created_at = mapped_column(TIMESTAMP(timezone=True), server_default=func.now())


class BlobsRepository(Repository[Blob]):
    Type = Blob

    def count_orphans(self):
        MH = aliased(Message, name="mh")
        MB = aliased(Message, name="mb")
        subquery = (
            select(Blob.id, func.count(MH.id) + func.count(MB.id))
            .select_from(Blob)
            .outerjoin(MH, MH.headers == Blob.id)
            .outerjoin(MB, MB.body == Blob.id)
            .group_by(Blob.id)
            .subquery()
        )
        query = select(func.count(subquery.c.id)).where(subquery.c[1] == 0)
        return query

    def delete_orphans(self):
        MH = aliased(Message, name="mh")
        MB = aliased(Message, name="mb")
        subquery = (
            select(Blob.id, func.count(MH.id) + func.count(MB.id))
            .select_from(Blob)
            .outerjoin(MH, MH.headers == Blob.id)
            .outerjoin(MB, MB.body == Blob.id)
            .group_by(Blob.id)
            .subquery()
        )
        query = select(subquery.c.id).where(subquery.c[1] == 0)
        return delete(Blob).where(Blob.id.in_(query))

    @with_session
    async def create(self, values: dict | BlobModel, /, *, session):
        if isinstance(values, BlobModel):
            values = dict(
                id=values.id,
                data=values.data,
                content_type=values.content_type,
            )
        return await super().create(values, session=session)



from typing import Callable, Optional, Self

from whistle import Event

from harp import get_logger
from harp.http import BaseHttpMessage, HttpError, HttpResponse
from harp.http.requests import HttpRequest
from harp.models import Transaction
from harp.views.json import serialize

logger = get_logger(__name__)

#: Event fired when a transaction is created.
EVENT_TRANSACTION_STARTED = "proxy.transaction.started"

#: Event fired when a message is sent to a transaction (either request or response).
EVENT_TRANSACTION_MESSAGE = "proxy.transaction.message"

#: Event fired when a transaction is finished, before the response is sent back to the caller.
EVENT_TRANSACTION_ENDED = "proxy.transaction.ended"

#: Event fired when an incoming request is ready to be filtered, for example by the rules application.
EVENT_FILTER_PROXY_REQUEST = "proxy.filter.request"

#: Event fired when an outgoing response is ready to be filtered, for example by the rules application.
EVENT_FILTER_PROXY_RESPONSE = "proxy.filter.response"

#: An error happened.
EVENT_PROXY_ERROR = "proxy.error"


class ProxyFilterEvent(Event):
    def __init__(
        self,
        endpoint: str,
        /,
        *,
        request: HttpRequest,
        response: Optional[HttpResponse] = None,
        transaction: Optional[Transaction] = None,
    ):
        self.endpoint = endpoint
        self.request = request
        self.response = response
        self.transaction = transaction

    def set_response(self, response: HttpResponse):
        if response is not None and not isinstance(response, HttpResponse):
            raise ValueError(
                f"Response must be an instance of HttpResponse, got {response!r} ({type(response).__module__}.{type(response).__qualname__})."
            )
        self.response = response

    def _get_rule_name(self) -> str | None:
        if not self.name:
            return None
        return "on_" + self.name.rsplit(".", 1)[-1]

    @property
    def criteria(self):
        return self.endpoint, str(self.request), self._get_rule_name()

    def create_execution_context(self):
        return {
            "logger": logger,
            "rule": self._get_rule_name(),
            "endpoint": self.endpoint,
            "request": self.request,
            "response": self.response,
            "stop_propagation": self.stop_propagation,
            "transaction": self.transaction,
        }

    def execute_script(self, script: Callable):
        context = self.create_execution_context()
        script(context)
        if context["response"] != self.response:
            if context["response"] is None:
                self.set_response(None)
            else:
                self.update(context["response"])
        return context

    def update(self, mixed: Optional[dict | HttpResponse | Self]) -> Self:
        if mixed is None:
            return self

        if isinstance(mixed, ProxyFilterEvent):
            return mixed

        if isinstance(mixed, dict):
            mixed = HttpResponse(serialize(mixed), content_type="application/json")

        if not isinstance(mixed, HttpResponse):
            raise ValueError(
                f"Response must be an instance of HttpResponse or a dict, got {mixed!r} ({type(mixed).__module__}.{type(mixed).__qualname__})."
            )

        self.set_response(mixed)
        return self


class TransactionEvent(Event):
    def __init__(self, transaction: Transaction):
        self.transaction = transaction


class HttpMessageEvent(TransactionEvent):
    def __init__(self, transaction: Transaction, message: BaseHttpMessage):
        super().__init__(transaction)
        self.message = message


class ProxyErrorEvent(TransactionEvent):
    def __init__(self, transaction: Transaction, error: HttpError):
        super().__init__(transaction)
        self.error = error



import os
from typing import cast

from click import BaseCommand

from harp.commandline import cookiecutters as templates
from harp.utils.commandline import check_packages, click


@click.command("create", short_help="Creates a templated directory using cookiecutter (project...).")
@click.argument("template", type=click.Choice(["project"], case_sensitive=False))
def create(template):
    """Creates a new project using cookiecutter."""

    if not check_packages("cookiecutter"):
        raise click.UsageError(
            "You need to install cookiecutter to use this command (or use the `dev` extra, for example using "
            "`pip install harp-proxy[dev]` or `poetry install -E dev`)."
        )

    from cookiecutter.main import cookiecutter

    cookiecutter(os.path.join(templates.__path__[0], template))


create = cast(BaseCommand, create)



import time
import traceback
from decimal import Decimal

import orjson
from pydantic import AnyUrl
from whistle import IAsyncEventDispatcher

from harp.asgi.events import EVENT_CORE_VIEW, ViewEvent
from harp.http import HttpResponse

STRINGIFIABLES = (
    AnyUrl,
    Decimal,
)

try:
    from freezegun.api import FakeDatetime

    STRINGIFIABLES += (FakeDatetime,)
except ImportError:
    pass


LISTIFIABLES = (set,)


def default(obj):
    if isinstance(obj, time.struct_time):
        return time.strftime("%Y-%m-%d %H:%M:%S", obj)
    if isinstance(obj, STRINGIFIABLES):
        return str(obj)
    if isinstance(obj, LISTIFIABLES):
        return list(sorted(obj))
    raise TypeError


class json(dict):
    """
    This is a marked dict type that our view event listener will recognize and serialize to json in a response.

    Usage::

        def handler():
            return json({"foo": "bar"})

    """

    pass


def serialize(value):
    return orjson.dumps(
        value,
        option=orjson.OPT_NON_STR_KEYS | orjson.OPT_NAIVE_UTC,
        default=default,
    )


async def on_json_response(event: ViewEvent):
    if isinstance(event.value, json):
        content_type = "application/json"
        try:
            serialized = serialize(event.value)
            event.set_response(
                HttpResponse(serialized, status=200, content_type=content_type),
            )
        except TypeError as exc:
            event.set_response(
                HttpResponse(
                    orjson.dumps(
                        {
                            "error": "Cannot serialize response to json.",
                            "type": type(exc).__name__,
                            "message": str(exc),
                            "traceback": traceback.format_exc(),
                            "value": repr(event.value),
                        },
                    ),
                    status=500,
                    content_type=content_type,
                ),
            )


def register(dispatcher: IAsyncEventDispatcher):
    dispatcher.add_listener(EVENT_CORE_VIEW, on_json_response)



from functools import partial

import pytest
from alembic import command
from sqlalchemy.ext.asyncio import AsyncEngine

from conftest import DEFAULT_STORAGE_SETTINGS
from harp.utils.testing.databases import TEST_DATABASES
from harp_apps.storage.services import SqlStorage

from .engines.sql_engine import create_async_engine
from .services.blob_storages.redis import RedisBlobStorage
from .settings import StorageSettings
from .utils.migrations import create_alembic_config, do_migrate
from .utils.testing.redis import create_redis_client
from .utils.testing.sql import create_database_container_for, get_scoped_database_url


@pytest.fixture
async def redis_blob_storage():
    async with create_redis_client() as async_client:
        from harp_apps.storage.services.blob_storages.redis import RedisBlobStorage

        yield RedisBlobStorage(async_client)


@pytest.fixture
async def sql_blob_storage(sql_engine):
    from harp_apps.storage.services.blob_storages.sql import SqlBlobStorage

    yield SqlBlobStorage(sql_engine)


@pytest.fixture(params=["sql", "redis"])
async def blob_storage(request, sql_engine):
    if request.param == "sql":
        from harp_apps.storage.services.blob_storages.sql import SqlBlobStorage

        yield SqlBlobStorage(sql_engine)
    elif request.param == "redis":
        async with create_redis_client() as async_client:
            yield RedisBlobStorage(async_client)

    else:
        raise ValueError(f"Unsupported blob storage type: {request.param}")


@pytest.fixture(scope="session", params=TEST_DATABASES)
def database_url(request):
    dialect, image, driver = request.param.split("|")
    yield from create_database_container_for(dialect, image, driver)


@pytest.fixture
async def sql_engine(database_url, test_id) -> AsyncEngine:
    """
    Use a DMBS conncetion string to create a sqlalchemy async engine to a new isolated database that will gets an
    initial status/structure/schema by executing migrations on it, whatever the migration process is for the given
    URL (for example, sqlite will use a simple create_all strategy, while postgres or mysql will use alembic).

    This looks a bit overkill, but this is the simplest way to guarantee isolated and repeatable tests that depends on
    databases.

    The isolated database name will be based on the test_id fixture value, which will create an unique hash string for
    each test.

    """
    async with get_scoped_database_url(database_url, test_id) as scoped_database_url:
        engine = create_async_engine(scoped_database_url)
        try:
            # todo refactor (dup in sqlstorage.initialize / sqlstorage._run_migrations)
            alembic_cfg = create_alembic_config(engine.url.render_as_string(hide_password=False))
            migrator = partial(command.upgrade, alembic_cfg, "head")
            await do_migrate(engine, migrator=migrator)
            yield engine
        finally:
            await engine.dispose()


@pytest.fixture
async def sql_storage(sql_engine, blob_storage) -> SqlStorage:
    storage = SqlStorage(
        sql_engine,
        blob_storage=blob_storage,
        settings=StorageSettings(**(DEFAULT_STORAGE_SETTINGS | {"url": sql_engine.url})),
    )
    # migrations are disabled because already done by the sql_engine fixture, but we still need to do additional
    # initialization steps.
    await storage.initialize()
    yield storage



from typing import Optional, Protocol


class NotificationSender(Protocol):
    async def send_notification(
        self,
        method: Optional[str],
        url: Optional[str],
        status_code: int,
        message: str,
        transaction_id: Optional[str],
    ) -> None: ...



import asyncio
import os

from asgi_middleware_static_file import ASGIMiddlewareStaticFile
from asgiref.typing import ASGISendCallable
from http_router import NotFoundError, Router
from httpx import AsyncClient

from harp import get_logger
from harp.controllers import RoutingController
from harp.errors import ConfigurationError
from harp.http import AlreadyHandledHttpResponse, HttpRequest, HttpResponse
from harp_apps import dashboard
from harp_apps.proxy.controllers import HttpProxyController
from harp_apps.proxy.settings.remote import Remote
from harp_apps.storage.types import IStorage

from ..settings import DashboardSettings
from ..settings.auth import BasicAuthSettings

logger = get_logger(__name__)


class DashboardController(RoutingController):
    name = "ui"

    storage: IStorage
    settings: DashboardSettings
    http_client: AsyncClient

    _ui_static_middleware = None
    _ui_devserver_proxy_controller = None

    #: Static directory to look for pre-built assets.
    static_build_path = os.path.realpath(os.path.join(os.path.dirname(os.path.abspath(dashboard.__file__)), "web"))

    def __init__(
        self,
        storage: IStorage,
        settings: DashboardSettings,
        http_client: AsyncClient,
        router: Router = None,
    ):
        super().__init__(router=router, handle_errors=False)

        # context for usage in handlers
        self.http_client = http_client
        self.storage = storage
        self.settings = settings

        # create users if they don't exist
        if isinstance(self.settings.auth, BasicAuthSettings):
            asyncio.create_task(self.storage.create_users_once_ready(self.settings.auth.users))

        if self.settings.enable_ui:
            self._initialize_ui()

    def _initialize_ui(self):
        # controllers for delegating requests
        if self.settings.devserver.enabled and self.settings.devserver.port:
            self._ui_devserver_proxy_controller = self._create_ui_devserver_proxy_controller(
                port=self.settings.devserver.port
            )

        # if no devserver is configured, we may need to serve static files
        if not self._ui_devserver_proxy_controller:
            if os.path.exists(self.static_build_path):
                self._ui_static_middleware = ASGIMiddlewareStaticFile(None, "", [self.static_build_path])

        # if no devserver is configured and no static files are found, we can't serve the dashboard
        if not self._ui_static_middleware and not self._ui_devserver_proxy_controller:
            raise ConfigurationError(
                "Dashboard controller could not initiate because it got neither compiled assets nor a devserver "
                "configuration. If you're using an editable install, make sure to run `make build-frontend` in the "
                "target working copy, or start a frontend development server."
            )

    def __repr__(self):
        features = {
            "api": True,
            "devserver": bool(self._ui_devserver_proxy_controller),
            "static": bool(self._ui_static_middleware),
        }
        return f"{type(self).__name__}({'+'.join(f for f in features if features[f])})"

    def _create_ui_devserver_proxy_controller(self, *, port):
        return HttpProxyController(
            Remote.from_settings_dict(
                {
                    "endpoints": [{"url": f"http://localhost:{port}/"}],
                    "liveness": {"type": "ignore"},
                }
            ),
            http_client=self.http_client,
            logging=False,
            name="dashboard-devserver",
        )

    async def __call__(self, request: HttpRequest, asgi_send: ASGISendCallable, *, transaction_id=None) -> HttpResponse:
        request.extensions.setdefault("user", None)

        if self.settings.auth:
            current_auth = request.basic_auth

            if current_auth:
                request.extensions["user"] = self.settings.auth.check(current_auth[0], current_auth[1])

            if not request.extensions["user"]:
                return HttpResponse(
                    b"Unauthorized",
                    status=401,
                    headers={"WWW-Authenticate": 'Basic realm="Harp Dashboard"'},
                    content_type="text/plain",
                )

        # Is this a prebuilt static asset? Can we serve it using our middleware? This wil need refactoring, as it ties
        # the controller implementation to the underlying webserver protocol implementation, but for now, it works.
        if (
            self._ui_static_middleware
            and not request.path.startswith("/api/")
            and hasattr(request._impl, "asgi_receive")
        ):
            # XXX todo fix
            await self._ui_static_middleware(
                {
                    "type": "http",
                    "path": request.path if "." in request.path else "/index.html",
                    "method": request.method,
                },
                request._impl.asgi_receive,
                asgi_send,
            )
            return AlreadyHandledHttpResponse()

        try:
            return await super().__call__(request)
        except NotFoundError:
            if self._ui_devserver_proxy_controller:
                return await self._ui_devserver_proxy_controller(request)

        return HttpResponse("Not found.", status=404, content_type="text/plain")



from hishel import HEURISTICALLY_CACHEABLE_STATUS_CODES

from harp.config import Configurable, Service


class CacheSettings(Configurable):
    #: Global cache flag, set to false to disable caching.
    enabled: bool = True

    #: Cache transport to use for the client. This is usually a hishel.AsyncCacheTransport (or subclass) instance.
    transport: Service = Service(type="hishel.AsyncCacheTransport")

    controller: Service = Service(
        type="hishel.Controller",
        arguments={
            "allow_heuristics": False,
            "allow_stale": False,
            "cacheable_methods": ["GET", "HEAD"],
            "cacheable_status_codes": list(HEURISTICALLY_CACHEABLE_STATUS_CODES),
        },
    )

    storage: Service = Service(
        base="hishel.AsyncBaseStorage",
        type="harp_apps.http_client.contrib.hishel.storages.AsyncStorage",
        arguments={
            "ttl": None,
            "check_ttl_every": 60.0,
        },
    )



from typing import AsyncIterator, cast

from asgiref.typing import ASGIReceiveCallable
from httpx import AsyncByteStream


class AsyncStreamFromAsgiReceive(AsyncByteStream):
    def __init__(self, asgi_receive: ASGIReceiveCallable) -> None:
        self.asgi_receive = asgi_receive
        self._closed = False

    async def __aiter__(self) -> AsyncIterator[bytes]:
        if self._closed:
            raise RuntimeError("Whole stream has already been read.")
        while not self._closed:
            message = await self.asgi_receive()
            self._closed = not message.get("more_body", False)
            yield cast(bytes, message.get("body", b""))

    async def aclose(self) -> None:
        self._closed = True



import typing as tp
from datetime import datetime

import yaml
from hishel._async._storages import StoredResponse
from hishel._serializers import KNOWN_REQUEST_EXTENSIONS, KNOWN_RESPONSE_EXTENSIONS, Metadata
from hishel._utils import normalized_url
from httpcore import Request, Response

from harp.models import Blob
from harp_apps.http_client.contrib.hishel.utils import (
    prepare_headers_for_deserialization,
    prepare_headers_for_serialization,
)
from harp_apps.storage.types import IBlobStorage


class SerializedRequest(tp.TypedDict):
    method: str
    url: str
    headers: str
    varying: dict[str, str]
    extensions: dict[str, str]


class SerializedResponse(tp.TypedDict):
    status: int
    headers: str
    varying: dict[str, str]
    body: str
    extensions: dict[str, str]


def _ensure_datestring(date: datetime | str):
    if isinstance(date, datetime):
        date = date.strftime("%a, %d %b %Y %H:%M:%S GMT")
    return date


class AsyncStorageAdapter:
    def __init__(self, storage: IBlobStorage):
        self.storage = storage

    async def store(self, key, /, *, response: Response, request: Request, metadata: Metadata) -> Blob:
        serialized_request = await self._serialize_request(request)
        serialized_response = await self._serialize_response(response)

        return await self._store_cache_meta(
            key,
            metadata=metadata,
            request=serialized_request,
            response=serialized_response,
        )

    async def retrieve(self, key: str) -> tp.Optional[StoredResponse]:
        # todo remove expired cache ?

        cached = await self.storage.get(key)
        if not cached:
            return None

        _metadata, _request, _response = await self._decode(cached)

        response = await self._unserialize_response(_response)
        request = await self._unserialize_request(_request)

        return (
            response,
            request,
            Metadata(
                cache_key=_metadata["cache_key"],
                created_at=datetime.strptime(_metadata["created_at"], "%a, %d %b %Y %H:%M:%S GMT"),
                number_of_uses=_metadata["number_of_uses"],
            ),
        )

    async def update_metadata_or_save(
        self, key: str, /, *, response: Response, request: Request, metadata: Metadata
    ) -> Blob:
        cached = await self.storage.get(key)
        if not cached:
            return await self.store(key, response=response, request=request, metadata=metadata)

        old_metadata, request_data, response_data = await self._decode(cached)
        await self._store_cache_meta(key, request=request_data, response=response_data, metadata=metadata)

    async def _decode(self, cached):
        cached = yaml.safe_load(cached.data.decode())
        request_data, response_data, raw_metadata = (
            cached["request"],
            cached["response"],
            cached["metadata"],
        )
        return raw_metadata, request_data, response_data

    async def _store_cache_meta(
        self,
        key,
        /,
        *,
        metadata: Metadata,
        request: SerializedRequest,
        response: SerializedResponse,
    ):
        # This is a special case where we don't want this to be content adressable. This is probably not very good, but
        # with hishel's current design, it's the only decent way to make it work that we found. Maybe we want to change
        # the key-value store in the future to be able to contain content addressable and unadressable data, even maybe
        # namespaced/typed data (although we hack "content-type to do it, for now).
        return await self.storage.force_put(
            Blob(
                id=key,
                data=yaml.safe_dump(
                    {
                        "request": request,
                        "response": response,
                        "metadata": {
                            "cache_key": metadata["cache_key"],
                            "number_of_uses": metadata["number_of_uses"],
                            "created_at": _ensure_datestring(metadata["created_at"]),
                        },
                    },
                    sort_keys=False,
                ).encode(),
                content_type="cache/meta",
            ),
        )

    async def _serialize_request(self, request: Request) -> SerializedRequest:
        headers, varying_headers, metadata = prepare_headers_for_serialization(request.headers)
        headers = await self.storage.put(Blob.from_data(headers, content_type="http/headers"))
        return {
            "method": request.method.decode("ascii"),
            "url": normalized_url(request.url),
            "headers": headers.id,
            "varying": varying_headers,
            "extensions": {key: value for key, value in request.extensions.items() if key in KNOWN_REQUEST_EXTENSIONS},
        }

    async def _unserialize_request(self, data: SerializedRequest) -> Request:
        headers = await self.storage.get(data["headers"])

        return Request(
            method=data["method"],
            url=data["url"],
            headers=prepare_headers_for_deserialization(headers.data, varying=data.get("varying") or {}),
            extensions=data.get("extensions") or {},
        )

    async def _serialize_response(self, response: Response) -> SerializedResponse:
        headers, varying_headers, metadata = prepare_headers_for_serialization(
            response.headers,
            varying=(
                b"date",
                b"content-length",
            ),
        )
        headers = await self.storage.put(Blob.from_data(headers, content_type="http/headers"))
        body = await self.storage.put(
            Blob.from_data(
                response.content,
                content_type=metadata.get("content-type") or "application/octet-stream",
            )
        )
        return {
            "status": response.status,
            "headers": headers.id,
            "varying": varying_headers,
            "body": body.id,
            "extensions": {
                key: value.decode("ascii")
                for key, value in response.extensions.items()
                if key in KNOWN_RESPONSE_EXTENSIONS
            },
        }

    async def _unserialize_response(self, data: SerializedResponse) -> Response:
        headers = await self.storage.get(data["headers"])
        body = await self.storage.get(data["body"])

        return Response(
            status=data["status"],
            headers=prepare_headers_for_deserialization(headers.data, varying=data.get("varying") or {}),
            content=body.data,
            extensions={key: value.encode() for key, value in (data.get("extensions") or {}).items()},
        )



"""
The Controllers (:mod:`harp.controllers`) package contains the routing and controller logic for the Harp framework.

You probably do not want to use this package unless you're writing an extension for Harp.

Example usage:

.. code-block:: python

    from harp.controllers import RoutingController, GetHandler


    class MyController(RoutingController):
        @GetHandler("/")
        async def index(self):
            return "Hello, world!"

Contents
--------

"""

from .default import dump_request_controller, not_found_controller
from .resolvers import DefaultControllerResolver, ProxyControllerResolver
from .routing import (
    AnyMethodHandler,
    ConnectHandler,
    DeleteHandler,
    GetHandler,
    HeadHandler,
    OptionsHandler,
    PatchHandler,
    PostHandler,
    PutHandler,
    RouteHandler,
    RouterPrefix,
    RoutingController,
    TraceHandler,
)
from .typing import IControllerResolver

__title__ = "Controllers"

__all__ = [
    "AnyMethodHandler",
    "ConnectHandler",
    "IControllerResolver",
    "DefaultControllerResolver",
    "DeleteHandler",
    "GetHandler",
    "HeadHandler",
    "OptionsHandler",
    "PatchHandler",
    "PostHandler",
    "ProxyControllerResolver",
    "PutHandler",
    "RouteHandler",
    "RouterPrefix",
    "RoutingController",
    "TraceHandler",
    "dump_request_controller",
    "not_found_controller",
]



from typing import cast

from click import BaseCommand

from harp import run
from harp.commandline.options.server import CommonServerOptions, server_command
from harp.config.utils import get_configuration_builder_type


@server_command(
    short_help="Starts HARP server.",
    help="""Starts HARP server, using the provided configuration. This is the main process and will be the only process
    you need on a live server, it will serve both the proxy ports and the compiled frontend assets (dashboard).""",
)
def server(**kwargs):
    from prometheus_client import Enum

    _info = Enum(
        "harp",
        "HARP status information.",
        states=["setup", "up", "teardown", "down"],
    )
    _info.state("setup")

    configration_builder = get_configuration_builder_type()

    builder = configration_builder.from_commandline_options(CommonServerOptions(**kwargs))

    _info.state("up")

    try:
        return run(builder)
    finally:
        _info.state("down")


server = cast(BaseCommand, server)



from importlib.machinery import ModuleSpec
from types import ModuleType
from unittest.mock import patch

import orjson

from harp.config import Application, ApplicationsRegistry, ConfigurationBuilder
from harp.config.asdict import asdict
from harp.utils.identifiers import is_valid_dotted_identifier


def test_is_valid_extension_name():
    # empty
    assert not is_valid_dotted_identifier("")

    # usual
    assert is_valid_dotted_identifier("foo.bar")
    assert is_valid_dotted_identifier("toto")

    # empty part
    assert not is_valid_dotted_identifier(".foo")
    assert not is_valid_dotted_identifier("foo..bar")
    assert not is_valid_dotted_identifier("bar.")

    # dashed/underscored
    assert is_valid_dotted_identifier("o_o")
    assert not is_valid_dotted_identifier("o-o")
    assert is_valid_dotted_identifier("_mazette")
    assert is_valid_dotted_identifier("raise_")


class ApplicationsRegistryMock(ApplicationsRegistry):
    def add_mock(self, name, impl):
        self._applications[name] = impl


def test_add_application(snapshot):
    builder = ConfigurationBuilder(use_default_applications=False)
    builder.applications.add("storage")

    assert len(builder.applications) == 1

    settings = builder()

    serialized = orjson.dumps(asdict(settings))
    assert serialized == snapshot

    new_builder = ConfigurationBuilder.from_bytes(serialized)
    assert asdict(new_builder()) == asdict(settings)

    module = ModuleType("foo.bar")
    module.__spec__ = ModuleSpec(name="foo.bar", loader=None)

    app_module = ModuleType("foo.bar.__app__")
    app_module.application = Application()
    app_module.__spec__ = ModuleSpec(name="foo.bar.__app__", loader=None)

    with patch.dict("sys.modules", {"foo.bar": module, "foo.bar.__app__": app_module}):
        new_builder.applications.add("foo.bar")
        new_settings = new_builder()

    assert asdict(new_settings) != asdict(settings)

    assert asdict(new_settings) == snapshot
    assert orjson.dumps(asdict(new_settings)) == snapshot



from typing import Annotated, Union

from pydantic import Discriminator

from .ignore import IgnoreLiveness, IgnoreLivenessSettings
from .inherit import InheritLiveness, InheritLivenessSettings
from .leaky_bucket import LeakyBucketLiveness, LeakyBucketLivenessSettings
from .naive import NaiveLiveness, NaiveLivenessSettings

LivenessSettings = Annotated[
    Union[
        IgnoreLivenessSettings,
        InheritLivenessSettings,
        LeakyBucketLivenessSettings,
        NaiveLivenessSettings,
    ],
    Discriminator("type"),
]

Liveness = Union[
    IgnoreLiveness,
    InheritLiveness,
    LeakyBucketLiveness,
    NaiveLiveness,
]



import time
import typing as tp
from datetime import datetime, timezone

from hishel import AsyncBaseStorage
from hishel._async._storages import StoredResponse
from hishel._serializers import Metadata
from httpcore import Request, Response

from harp_apps.proxy.controllers import logger
from harp_apps.storage.types import IBlobStorage

from .adapters import AsyncStorageAdapter

HEADERS_ENCODING = "iso-8859-1"


class AsyncStorage(AsyncBaseStorage):
    def __init__(
        self,
        storage: IBlobStorage,
        ttl: tp.Optional[tp.Union[int, float]] = None,
        check_ttl_every: tp.Union[int, float] = 60,
    ):
        super().__init__(serializer=None, ttl=ttl)

        self._check_ttl_every = check_ttl_every
        self._last_cleaned = time.monotonic()
        self._impl = AsyncStorageAdapter(storage)
        self._storage = storage

    async def store(
        self,
        key: str,
        response: Response,
        request: Request,
        metadata: Metadata | None = None,
    ) -> None:
        # XXX this looks like the wrong place to do it, but hishel depends on this behaviour. Let's mimic the other
        #  storages, for now.
        metadata = metadata or Metadata(cache_key=key, created_at=datetime.now(timezone.utc), number_of_uses=0)

        await self._impl.store(
            key,
            request=request,
            response=response,
            metadata=metadata,
        )

    async def update_metadata(self, key: str, response: Response, request: Request, metadata: Metadata) -> None:
        await self._impl.update_metadata_or_save(
            key,
            request=request,
            response=response,
            metadata=metadata,
        )

    async def retrieve(self, key: str) -> tp.Optional[StoredResponse]:
        try:
            return await self._impl.retrieve(key)
        except Exception:
            logger.exception("Failed to retrieve cache")

    async def aclose(self) -> None:
        return



from typing import Optional

from harp.config import ConfigurationBuilder
from harp.config.builders.system import System


class BaseTestDefaultsWith:
    default_applications = ("http_client",)

    async def create_system(self, values: Optional[dict] = None, /, *, applications=None) -> System:
        applications = applications or self.default_applications
        config = ConfigurationBuilder({"applications": applications}, use_default_applications=False)
        if values is not None:
            config.add_values(values)
        return await config.abuild_system()



from datetime import UTC, datetime, timedelta
from typing import List

from harp_apps.storage.types import TransactionsGroupedByTimeBucket


def get_start_datetime_from_range(range: str | None, *, now=None) -> datetime | None:
    """
    Generate a list of datetime objects from a range string.
    """
    if not range:
        return None
    now = now or datetime.now(UTC)
    if range == "1h":
        start_datetime = now - timedelta(hours=1)
    elif range == "24h":
        start_datetime = now - timedelta(hours=24)
    elif range == "7d":
        start_datetime = now - timedelta(days=7)
    elif range == "1m":
        start_datetime = now - timedelta(days=30)
    elif range == "1y":
        start_datetime = now - timedelta(days=365)
    else:
        raise ValueError(f"Unknown range: {range}")
    return start_datetime


def _truncate_datetime_for_time_bucket(dt: datetime, time_bucket: str = "day") -> datetime:
    """
    Truncate a datetime object to the given time bucket.
    """
    if time_bucket == "day":
        return dt.replace(hour=0, minute=0, second=0, microsecond=0)
    elif time_bucket == "hour":
        return dt.replace(minute=0, second=0, microsecond=0)
    elif time_bucket == "minute":
        return dt.replace(second=0, microsecond=0)
    else:
        raise ValueError(f"Unknown time bucket: {time_bucket}")


def generate_continuous_time_range(
    discontinuous_transactions: List[TransactionsGroupedByTimeBucket],
    start_datetime=None,
    time_bucket: str = "day",
) -> List[TransactionsGroupedByTimeBucket]:
    """
    Generate a list of datetime objects from a range string.
    """
    # Generate a list of datetime objects from a range string
    start_datetime = start_datetime if start_datetime else discontinuous_transactions[0]["datetime"]
    start_datetime = _truncate_datetime_for_time_bucket(start_datetime, time_bucket)

    end_datetime = datetime.now(UTC)
    delta = timedelta(
        days=1 if time_bucket == "day" else 0,
        hours=1 if time_bucket == "hour" else 0,
        minutes=1 if time_bucket == "minute" else 0,
    )

    datetime_range = [start_datetime + i * delta for i in range(int((end_datetime - start_datetime) / delta) + 1)]

    # Fill missing data points
    continuous_transactions = []
    for t in datetime_range:
        if t in [d["datetime"] for d in discontinuous_transactions]:
            continuous_transactions.append([d for d in discontinuous_transactions if d["datetime"] == t][0])
        else:
            continuous_transactions.append(
                {
                    "datetime": t,
                    "count": None,
                    "errors": None,
                    "cached": 0,
                    "meanDuration": None,
                    "meanTpdex": None,
                }
            )
    return continuous_transactions



from contextlib import asynccontextmanager
from typing import TYPE_CHECKING
from urllib.parse import urljoin

import httpx
from pydantic import Field
from pyheck import shouty_snake

from harp import get_logger
from harp.config import Configurable, Stateful

if TYPE_CHECKING:
    from harp_apps.proxy.settings.remote.endpoint import RemoteEndpoint

logger = get_logger(__name__)


class RemoteProbeSettings(Configurable):
    """
    A ``HttpProbe`` is a health check that can be used to check the health of a remote's endpoints. It is used as the
    configuration parser for ``proxy.endpoints[].remote.probe`` settings.

    .. code-block:: yaml

        type: http
        method: GET
        path: /health
        headers:
          x-purpose: "health probe"
        timeout: 5.0
    """

    method: str = "GET"
    path: str = "/"
    headers: dict = Field(default_factory=dict)
    interval: float = 10.0
    timeout: float = 10.0
    verify: bool = True


class RemoteProbe(Stateful[RemoteProbeSettings]):
    """Stateful version of a probe definition."""

    @asynccontextmanager
    async def async_client(self):
        async with httpx.AsyncClient(verify=self.settings.verify) as client:
            yield client

    async def check(self, client: httpx.AsyncClient, endpoint: "RemoteEndpoint"):
        probe_url = urljoin(str(endpoint.settings.url), self.settings.path)
        response = None
        try:
            response = await client.request(
                self.settings.method,
                probe_url,
                headers=self.settings.headers,
                timeout=float(self.settings.timeout),
            )
            if 200 <= response.status_code < 400:
                return endpoint.success()
            else:
                failure = f"PROBE_HTTP_{response.status_code}"
                logger.warning(f"Probe failure: {probe_url} -> {failure}")
                return endpoint.failure(failure)
        except Exception as exc:
            failure = "PROBE_" + shouty_snake(type(exc).__name__)
            logger.error(f"Probe failure: {probe_url} -> {failure}")
            return endpoint.failure(failure)
        finally:
            if response:
                logger.debug(f"Probe request: {probe_url} -> {response.status_code}")



from enum import Enum


class TimeBucket(Enum):
    YEAR = "year"
    MONTH = "month"
    WEEK = "week"
    DAY = "day"
    HOUR = "hour"
    MINUTE = "minute"


# Transaction markers that control storage behavior
SKIP_STORAGE = "skip-storage"
SKIP_REQUEST_STORAGE = "skip-request-storage"
SKIP_RESPONSE_STORAGE = "skip-response-storage"
SKIP_REQUEST_BODY_STORAGE = "skip-request-body-storage"
SKIP_RESPONSE_BODY_STORAGE = "skip-response-body-storage"
SKIP_REQUEST_HEADERS_STORAGE = "skip-request-headers-storage"
SKIP_RESPONSE_HEADERS_STORAGE = "skip-response-headers-storage"



import asyncio
from typing import cast

from prometheus_client import Gauge

from harp import get_logger
from harp_apps.storage.services import SqlStorage

from ..storage.models.base import with_session
from ..storage.types import IBlobStorage, IStorage
from .settings import OLD_AFTER, PERIOD

logger = get_logger(__name__)

PROMETHEUS_GAUGES = {
    "storage.transactions": Gauge("storage_transactions", "Transactions currently in storage."),
    "storage.messages": Gauge("storage_messages", "Messages currently in storage."),
    "storage.blobs": Gauge("storage_blobs", "Blob objects currently in storage."),
    "storage.blobs.orphans": Gauge("storage_blobs_orphans", "Orphan blobs currently in storage."),
}


class JanitorWorker:
    def __init__(self, storage: IStorage, blob_storage: IBlobStorage):
        self.storage: SqlStorage = cast(SqlStorage, storage)
        self.blob_storage: IBlobStorage = blob_storage

        self.running = False
        self.session_factory = self.storage.session_factory
        self._running_lock = asyncio.Lock()

    def stop(self):
        """
        Mark the loop for termination.
        """
        self.running = False

    async def run(self):
        """
        Once dependencies are ready, start the main loop (basically, run the `loop()` every PERIOD seconds), until
        `stop()` is called.
        """
        # do not start before storage is ready
        async with self._running_lock:
            self.running = True

            while self.running:
                try:
                    await self.loop()
                except Exception as exc:
                    logger.exception(exc)
                await asyncio.sleep(PERIOD)

    async def loop(self):
        """
        One iteration of the janitor loop.
        """

        # Delete old transactions
        result = await self.delete_old_transactions()
        if result.rowcount:
            logger.debug("🧹 Deleted %d old transactions", result.rowcount)

        await self.delete_orphan_blobs()

        # Compute and store stored objecg counts as metrics
        logger.debug("🧹 Compute and store metrics...")
        await self.compute_and_store_metrics()

    @with_session
    async def delete_old_transactions(self, /, *, session):
        """
        Remove transactions older than OLD_AFTER days. On correct database implementations (postgresql for example), it
        will cascade to related objects. On sqlite, there will be garbage left, but it's not a big deal.
        """

        # TODO to storage

        result = await session.execute(self.storage.transactions.delete_old(OLD_AFTER))
        await session.commit()
        return result

    @with_session
    async def delete_orphan_blobs(self, /, *, session):
        """
        Find and remove blobs that are not referenced anymore by any transaction.
        """

        count = None

        if self.blob_storage.type == "sql":
            result = await session.execute(self.storage.blobs.delete_orphans())
            await session.commit()
            count = result.rowcount if result.rowcount else 0
        elif self.blob_storage.type == "redis":
            pass
        else:
            pass

        if count is None:
            # The blob storage may need to clean orphans but no implementation is available
            logger.debug("🧹 DeleteOrphanBlobs[%s] Not implemented.", self.blob_storage.type)
        elif count is False:
            # The blob storage does not NEED to delete orphans (for example for a NullBlobStorage)
            pass
        else:
            logger.debug(
                "🧹 DeleteOrphanBlobs[%s] Removed %d blobs.",
                self.blob_storage.type,
                count,
            )

    @with_session
    async def compute_and_store_metrics(self, /, *, session):
        """
        Compute counts of objects in storage, and store them as metrics.
        """

        # TODO to storage

        await self.storage.metrics.insert_values(await self.compute_metrics(session))

    async def compute_metrics(self, session):
        values = {
            "storage.transactions": await self.do_count(session, "transactions"),
            "storage.messages": await self.do_count(session, "messages"),
            "storage.blobs": await self.do_count(session, "blobs"),
            "storage.blobs.orphans": await self.do_count(session, "blobs", method="count_orphans"),
        }

        for key, value in values.items():
            PROMETHEUS_GAUGES[key].set(value)

        return values

    async def do_count(self, session, name: str, /, *, method="count"):
        """
        Helper to count objects in storage, from different repositories and using different methods for building the
        actual query.

        :param session: sqlalchemy async session
        :param name: repository name (should be available from storage)
        :param method: method name to call on the repository to get the actual sqlalchemy query, default as "count"
        :return: integer
        """

        # TODO to storage
        return (
            await session.execute(
                getattr(
                    getattr(self.storage, name),
                    method,
                )()
            )
        ).scalar()



from asgiref.testing import ApplicationCommunicator

from .asgi import ASGICommunicator
from .http import HTTPCommunicator

__all__ = [
    "ASGICommunicator",
    "ApplicationCommunicator",
    "HTTPCommunicator",
]



from typing import Optional

from pydantic import Field

from harp.config import Configurable
from harp_apps.dashboard.settings.auth import BasicAuthSettings
from harp_apps.dashboard.settings.devserver import DevserverSettings


class DashboardSettings(Configurable):
    """Root settings for the dashboard application."""

    port: int = Field(
        4080,
        description="Port on which the dashboard application will be served.",
    )

    auth: Optional[BasicAuthSettings] = Field(
        None,
        discriminator="type",
        description="Authentication settings for the dashboard.",
    )

    devserver: Optional[DevserverSettings] = Field(
        default_factory=DevserverSettings,
        description="Development server settings, only useful for internal frontend development.",
    )

    enable_ui: bool = Field(
        True,
        description="DEPRECATED – Whether to enable the dashboard UI.",
    )

    public_url: Optional[str] = Field(
        None,
        description="Public URL of the dashboard application, used to generate absolute links, for example in notifications.",
    )



import os
import shlex
import sys
from pathlib import Path
from string import Template

import rich_click as click

from harp import ROOT_DIR, get_logger
from harp.utils.network import get_available_network_port

HARP_DASHBOARD_SERVICE = "harp:dashboard"
HARP_DOCS_SERVICE = "harp:docs"
HARP_SERVER_SERVICE = "harp:server"
HARP_UI_SERVICE = "harp:ui"

logger = get_logger(__name__)


def quote(x):
    return shlex.quote(str(x))


class HonchoManagerFactory:
    defaults = {HARP_DASHBOARD_SERVICE, HARP_SERVER_SERVICE}
    commands = {}
    frontend_source_path: str = str(Path(ROOT_DIR).joinpath("harp_apps/dashboard/frontend").resolve())

    def __init__(self, *, proxy_options=(), dashboard_devserver_port=None):
        self.ports = {HARP_DASHBOARD_SERVICE: dashboard_devserver_port or get_available_network_port()}
        self.proxy_ports = {}
        self.proxy_options = proxy_options
        self.cwds = {}

        # copy to allow changes on this instance only
        self.names = {
            HARP_DASHBOARD_SERVICE,
            HARP_SERVER_SERVICE,
            HARP_DOCS_SERVICE,
            HARP_UI_SERVICE,
        }
        self.commands = {**self.commands}

    def _get_dashboard_executable(self, processes):
        # todo make sure the frontend tools are available, in the right versions
        if not os.path.exists(self.frontend_source_path):
            # todo better guidance
            raise click.UsageError(
                "Dashboard's frontend dependencies are not installed.\nYour options are: run a production version "
                "(shortcut to come), install the dependencies (with `harp install-dev`), or do not run the dashboard."
            )

        host = "localhost"

        return (
            self.frontend_source_path,
            " ".join(
                [
                    "pnpm exec vite",
                    f"--host {host}",
                    f"--port {self.ports[HARP_DASHBOARD_SERVICE]}",
                    "--strictPort",
                    "--clearScreen false",
                ]
            ),
        )

    commands[HARP_DASHBOARD_SERVICE] = _get_dashboard_executable

    def _get_server_executable(self, processes):
        cmd = f"{sys.executable} -m harp server"
        proxy_options = list(self.proxy_options)

        if HARP_DASHBOARD_SERVICE in processes:
            proxy_options.append(f"--set dashboard.devserver.port={quote(self.ports[HARP_DASHBOARD_SERVICE])}")

        for _name, _port in self.proxy_ports.items():
            proxy_options.append(f"--endpoint {quote(_name)}={_port}:http://localhost:{self.ports[_name]}")

        if proxy_options:
            cmd += " " + " ".join(proxy_options)

        return (
            None,
            f'watchfiles --filter python "{cmd}" {os.path.join(ROOT_DIR, "harp")} {os.path.join(ROOT_DIR, "harp_apps")}',
        )

    commands[HARP_SERVER_SERVICE] = _get_server_executable

    def _get_docs_executable(self, processes):
        # todo add check available
        return (
            os.path.join(ROOT_DIR, "docs"),
            "poetry run sphinx-autobuild . _build/html",
        )

    commands[HARP_DOCS_SERVICE] = _get_docs_executable

    def _get_ui_executable(self, processes):
        # todo add check available
        return os.path.join(ROOT_DIR, "harp_apps/dashboard/frontend"), "pnpm ui:serve"

    commands[HARP_UI_SERVICE] = _get_ui_executable

    def build(self, processes, /, *, more_env=None):
        from honcho.manager import Manager
        from honcho.printer import Printer

        manager = Manager(Printer(sys.stdout))
        for name in processes:
            if name not in self.commands:
                raise ValueError(f"Unknown process: {name}")

            if callable(self.commands[name]):
                working_directory, command = self.commands[name](self, processes)
            else:
                working_directory, command = (
                    self.cwds.get(name, None),
                    self.commands[name],
                )
                if working_directory is None:
                    working_directory = os.getcwd()
                else:
                    working_directory = os.path.join(os.getcwd(), working_directory)

            e = os.environ.copy()
            more_env = more_env or {}
            manager.add_process(name, command, cwd=working_directory, env=e | more_env.get(name, {}))

            logger.debug(f"Added process {name}: {command}")

            # this hack will change the class impl at runtime for frontend process to avoid misleading log at start.
            if name == HARP_DASHBOARD_SERVICE:
                from harp.commandline.utils._hacks import QuietViteHonchoProcess

                manager._processes[name]["obj"].__class__ = QuietViteHonchoProcess

        return manager


def parse_server_subprocesses_options(server_subprocesses):
    # TODO uniformise with --endpoint to be in format name=port:cmd
    processes = {}
    for server_subprocess in server_subprocesses:
        try:
            _name, _port, _path, _command = server_subprocess.split(":", 3)
        except ValueError as exc:
            raise click.UsageError(
                "Invalid server subprocess configuration. Expected format: <name>:<port>:<path>:<command>."
            ) from exc
        _target_port = get_available_network_port()
        _command = Template(_command).safe_substitute(port=_target_port)
        processes[_name] = (_port, _path, _command, _target_port)

    return processes


def get_honcho_manager_factory_type():
    edition = os.environ.get("HARP_EDITION", "harp_apps")
    try:
        HonchoManagerFactoryType = __import__(edition, fromlist=["HonchoManagerFactory"]).HonchoManagerFactory
    except (ImportError, AttributeError):
        return HonchoManagerFactory
    return HonchoManagerFactoryType



from base64 import b64encode
from typing import cast
from unittest.mock import AsyncMock, Mock

from httpx import AsyncClient

from harp.config.asdict import asdict
from harp.http import HttpRequest
from harp.utils.bytes import ensure_bytes
from harp_apps.dashboard.controllers import DashboardController
from harp_apps.dashboard.settings import BasicAuthSettings, DashboardSettings
from harp_apps.dashboard.settings.auth import User
from harp_apps.storage.types import IStorage


async def test_controller_no_auth():
    controller = await _create_mock_controller()
    response = await controller(HttpRequest(), AsyncMock())
    assert response.status == 404


async def test_controller_auth_plaintext():
    controller = await _create_mock_controller(
        settings=DashboardSettings(
            auth=BasicAuthSettings(
                type="basic",
                algorithm="plaintext",
                users={"admin": User(password="admin")},
            ),
            enable_ui=False,
        )
    )
    assert asdict(controller.settings.auth) == {
        "type": "basic",
        "algorithm": "plaintext",
        "users": {"admin": {"password": "admin"}},
    }

    # no auth? no chance
    response = await controller(HttpRequest(), AsyncMock())
    assert response.status == 401

    # sir, yes sir
    response = await controller(HttpRequest(headers=_get_auth_headers("admin", "admin")), AsyncMock())
    assert response.status == 404

    # wrong password
    response = await controller(HttpRequest(headers=_get_auth_headers("admin", "wrong")), AsyncMock())
    assert response.status == 401


async def _create_mock_controller(settings: DashboardSettings = None):
    return DashboardController(
        storage=cast(IStorage, Mock(spec=IStorage)),
        settings=settings or DashboardSettings(enable_ui=False),
        http_client=Mock(spec=AsyncClient),
    )


def _get_auth_headers(username, password):
    return {"authorization": "basic " + b64encode(b":".join((ensure_bytes(username), ensure_bytes(password)))).decode()}



import respx
from httpx import Response

from harp.utils.testing.communicators import ASGICommunicator
from harp_apps.storage.types import IBlobStorage, IStorage
from harp_apps.storage.worker import StorageAsyncWorkerQueue

from .._base import BaseRulesFlowTest


class TestSkipBodyStorageMarkers(BaseRulesFlowTest):
    """Integration tests for skip body storage markers set through rules."""

    applications = ["http_client", "proxy", "rules", "storage"]

    @respx.mock
    async def test_skip_request_body_storage_via_rules(self, httpbin):
        """Test that skip-request-body-storage marker set in rules prevents request body storage."""
        respx.post(httpbin).mock(return_value=Response(200, content=b"upstream response"))

        # Configure rules to add the skip-request-body-storage marker
        system = await self.create_system(
            {
                "proxy": {"endpoints": [{"name": "httpbin", "port": 80, "url": httpbin}]},
                "rules": {
                    "*": {
                        "*": {
                            "on_request": 'transaction.markers.add("skip-request-body-storage")',
                        }
                    }
                },
            },
            mock=lambda ctx: None,  # dummy mock
        )

        client = ASGICommunicator(system.asgi_app)
        await client.asgi_lifespan_startup()

        # Make a POST request with a body
        await client.http_post("/", body=b"sensitive request data")

        # Wait for storage worker to process events
        worker = system.provider.get(StorageAsyncWorkerQueue)
        await worker.wait_until_empty()

        # Get storage instances from system
        sql_storage = system.provider.get(IStorage)
        blob_storage = system.provider.get(IBlobStorage)

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1

        # Find request message
        request_msg = next(msg for msg in transactions[0].messages if msg.kind == "request")
        response_msg = next(msg for msg in transactions[0].messages if msg.kind == "response")

        # Verify request body was NOT stored
        assert request_msg.body is None

        # Verify request headers WERE stored
        assert request_msg.headers is not None

        # Verify response body WAS stored (marker only affects request)
        assert response_msg.body is not None
        response_blob = await blob_storage.get(response_msg.body)
        assert response_blob.data == b"upstream response"

    @respx.mock
    async def test_skip_response_body_storage_via_rules(self, httpbin):
        """Test that skip-response-body-storage marker set in rules prevents response body storage."""
        respx.get(httpbin).mock(return_value=Response(200, content=b"sensitive response data"))

        # Configure rules to add the skip-response-body-storage marker
        system = await self.create_system(
            {
                "proxy": {"endpoints": [{"name": "httpbin", "port": 80, "url": httpbin}]},
                "rules": {
                    "*": {
                        "*": {
                            "on_response": 'transaction.markers.add("skip-response-body-storage")',
                        }
                    }
                },
            },
            mock=lambda ctx: None,
        )

        client = ASGICommunicator(system.asgi_app)
        await client.asgi_lifespan_startup()

        # Make a GET request
        await client.http_get("/")

        # Wait for storage worker to process events
        worker = system.provider.get(StorageAsyncWorkerQueue)
        await worker.wait_until_empty()

        # Get storage instances from system
        sql_storage = system.provider.get(IStorage)

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1

        # Find messages
        request_msg = next(msg for msg in transactions[0].messages if msg.kind == "request")
        response_msg = next(msg for msg in transactions[0].messages if msg.kind == "response")

        # Verify request body WAS stored (marker only affects response)
        assert request_msg.body is not None

        # Verify response body was NOT stored
        assert response_msg.body is None

        # Verify response headers WERE stored
        assert response_msg.headers is not None

    @respx.mock
    async def test_skip_both_bodies_via_rules(self, httpbin):
        """Test that both markers set in rules prevents both request and response body storage."""
        respx.post(httpbin).mock(return_value=Response(200, content=b"response data"))

        # Configure rules to add both markers
        system = await self.create_system(
            {
                "proxy": {"endpoints": [{"name": "httpbin", "port": 80, "url": httpbin}]},
                "rules": {
                    "*": {
                        "*": {
                            "on_request": 'transaction.markers.add("skip-request-body-storage")',
                            "on_response": 'transaction.markers.add("skip-response-body-storage")',
                        }
                    }
                },
            },
            mock=lambda ctx: None,
        )

        client = ASGICommunicator(system.asgi_app)
        await client.asgi_lifespan_startup()

        # Make a POST request with a body
        await client.http_post("/", body=b"request data")

        # Wait for storage worker to process events
        worker = system.provider.get(StorageAsyncWorkerQueue)
        await worker.wait_until_empty()

        # Get storage instances from system
        sql_storage = system.provider.get(IStorage)

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1

        # Find messages
        request_msg = next(msg for msg in transactions[0].messages if msg.kind == "request")
        response_msg = next(msg for msg in transactions[0].messages if msg.kind == "response")

        # Verify neither body was stored
        assert request_msg.body is None
        assert response_msg.body is None

        # Verify headers WERE still stored for both
        assert request_msg.headers is not None
        assert response_msg.headers is not None

    @respx.mock
    async def test_conditional_skip_based_on_endpoint(self, httpbin):
        """Test that markers can be conditionally applied based on endpoint/path."""
        respx.post(f"{httpbin}/upload").mock(return_value=Response(200, content=b"upload ok"))
        respx.post(f"{httpbin}/process").mock(return_value=Response(200, content=b"process ok"))

        # Configure rules to skip only for /upload endpoint
        system = await self.create_system(
            {
                "proxy": {"endpoints": [{"name": "httpbin", "port": 80, "url": httpbin}]},
                "rules": {
                    "*": {
                        "POST /upload": {
                            "on_request": 'transaction.markers.add("skip-request-body-storage")',
                        }
                    }
                },
            },
            mock=lambda ctx: None,
        )

        client = ASGICommunicator(system.asgi_app)
        await client.asgi_lifespan_startup()

        # Make request to /upload (should skip)
        await client.http_post("/upload", body=b"large file data")

        # Make request to /process (should NOT skip)
        await client.http_post("/process", body=b"process data")

        # Wait for storage worker to process events
        worker = system.provider.get(StorageAsyncWorkerQueue)
        await worker.wait_until_empty()

        # Get storage instances from system
        sql_storage = system.provider.get(IStorage)
        blob_storage = system.provider.get(IBlobStorage)

        # Verify both transactions were stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 2

        # Find transactions by their paths (order may vary)
        upload_transaction = next(t for t in transactions if any("upload" in msg.summary for msg in t.messages))
        process_transaction = next(t for t in transactions if any("process" in msg.summary for msg in t.messages))

        upload_request_msg = next(msg for msg in upload_transaction.messages if msg.kind == "request")
        process_request_msg = next(msg for msg in process_transaction.messages if msg.kind == "request")

        # Verify upload request body was NOT stored
        assert upload_request_msg.body is None

        # Verify process request body WAS stored
        assert process_request_msg.body is not None
        process_blob = await blob_storage.get(process_request_msg.body)
        assert process_blob.data == b"process data"

    @respx.mock
    async def test_skip_request_headers_storage_via_rules(self, httpbin):
        """Test that skip-request-headers-storage marker set in rules prevents request headers storage."""
        respx.post(httpbin).mock(return_value=Response(200, content=b"upstream response"))

        # Configure rules to add the skip-request-headers-storage marker
        system = await self.create_system(
            {
                "proxy": {"endpoints": [{"name": "httpbin", "port": 80, "url": httpbin}]},
                "rules": {
                    "*": {
                        "*": {
                            "on_request": 'transaction.markers.add("skip-request-headers-storage")',
                        }
                    }
                },
            },
            mock=lambda ctx: None,
        )

        client = ASGICommunicator(system.asgi_app)
        await client.asgi_lifespan_startup()

        # Make a POST request
        await client.http_post("/", body=b"request data")

        # Wait for storage worker to process events
        worker = system.provider.get(StorageAsyncWorkerQueue)
        await worker.wait_until_empty()

        # Get storage instances from system
        sql_storage = system.provider.get(IStorage)
        blob_storage = system.provider.get(IBlobStorage)

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1

        # Find request message
        request_msg = next(msg for msg in transactions[0].messages if msg.kind == "request")
        response_msg = next(msg for msg in transactions[0].messages if msg.kind == "response")

        # Verify request headers were NOT stored
        assert request_msg.headers is None

        # Verify request body WAS stored
        assert request_msg.body is not None
        request_blob = await blob_storage.get(request_msg.body)
        assert request_blob.data == b"request data"

        # Verify response headers WAS stored (marker only affects request)
        assert response_msg.headers is not None

    @respx.mock
    async def test_skip_request_storage_via_rules(self, httpbin):
        """Test that skip-request-storage marker set in rules prevents entire request message storage."""
        respx.post(httpbin).mock(return_value=Response(200, content=b"upstream response"))

        # Configure rules to add the skip-request-storage marker
        system = await self.create_system(
            {
                "proxy": {"endpoints": [{"name": "httpbin", "port": 80, "url": httpbin}]},
                "rules": {
                    "*": {
                        "*": {
                            "on_request": 'transaction.markers.add("skip-request-storage")',
                        }
                    }
                },
            },
            mock=lambda ctx: None,
        )

        client = ASGICommunicator(system.asgi_app)
        await client.asgi_lifespan_startup()

        # Make a POST request with a body
        await client.http_post("/", body=b"request data")

        # Wait for storage worker to process events
        worker = system.provider.get(StorageAsyncWorkerQueue)
        await worker.wait_until_empty()

        # Get storage instances from system
        sql_storage = system.provider.get(IStorage)
        blob_storage = system.provider.get(IBlobStorage)

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1

        # Verify only response message was stored (request was skipped)
        assert len(transactions[0].messages) == 1
        response_msg = transactions[0].messages[0]

        # Verify it's the response message
        assert response_msg.kind == "response"
        assert response_msg.headers is not None
        assert response_msg.body is not None
        response_blob = await blob_storage.get(response_msg.body)
        assert response_blob.data == b"upstream response"

    @respx.mock
    async def test_skip_response_storage_via_rules(self, httpbin):
        """Test that skip-response-storage marker set in rules prevents entire response message storage."""
        respx.post(httpbin).mock(return_value=Response(200, content=b"upstream response"))

        # Configure rules to add the skip-response-storage marker
        system = await self.create_system(
            {
                "proxy": {"endpoints": [{"name": "httpbin", "port": 80, "url": httpbin}]},
                "rules": {
                    "*": {
                        "*": {
                            "on_response": 'transaction.markers.add("skip-response-storage")',
                        }
                    }
                },
            },
            mock=lambda ctx: None,
        )

        client = ASGICommunicator(system.asgi_app)
        await client.asgi_lifespan_startup()

        # Make a POST request with a body
        await client.http_post("/", body=b"request data")

        # Wait for storage worker to process events
        worker = system.provider.get(StorageAsyncWorkerQueue)
        await worker.wait_until_empty()

        # Get storage instances from system
        sql_storage = system.provider.get(IStorage)
        blob_storage = system.provider.get(IBlobStorage)

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1

        # Verify only request message was stored (response was skipped)
        assert len(transactions[0].messages) == 1
        request_msg = transactions[0].messages[0]

        # Verify it's the request message
        assert request_msg.kind == "request"
        assert request_msg.headers is not None
        assert request_msg.body is not None
        request_blob = await blob_storage.get(request_msg.body)
        assert request_blob.data == b"request data"



from tempfile import NamedTemporaryFile
from unittest.mock import Mock

from harp_apps.rules.models.scripts import Script


def test_script_from_sources():
    script = Script("print('Hello, World!')")
    assert script.source == "print('Hello, World!')"

    print_mock = Mock()
    script({"print": print_mock})

    assert print_mock.called
    assert print_mock.call_count == 1


def test_script_from_file():
    with NamedTemporaryFile("w", delete=False) as f:
        f.write("print('Hello, World!')")
        f.flush()

    script = Script.from_file(f.name)
    assert script.source == "print('Hello, World!')"

    print_mock = Mock()
    script({"print": print_mock})

    assert print_mock.called
    assert print_mock.call_count == 1



from functools import cached_property
from inspect import signature
from typing import List, Optional, Type

from rodi import CannotResolveTypeException, ResolutionContext, ServiceLifeStyle

from harp.config.configurables.service import LazyService
from harp.utils.packages import import_string

from .containers import Container
from .models import ServiceDefinition
from .providers import ServiceProvider
from .references import LazyServiceReference


class ServiceResolver:
    container: Container
    service: ServiceDefinition
    lifestyle: ServiceLifeStyle

    def __init__(
        self,
        container: Container,
        service: ServiceDefinition,
        lifestyle: Optional[ServiceLifeStyle] = None,
    ):
        self.container = container
        self.service = service
        self.lifestyle = (
            getattr(ServiceLifeStyle, (service.lifestyle or "singleton").upper()) if lifestyle is None else lifestyle
        )

    def _get_resolver(self, desired_type: str | list, context: ResolutionContext):
        # we use a list to support fallbacks
        if not isinstance(desired_type, List):
            desired_type = [desired_type]

        for _type in desired_type:
            # first, resolve aliases
            if _type in self.container._exact_aliases:
                _type = self.container._exact_aliases[_type]
            elif _type in self.container._aliases:
                # XXX this looks half implemented in rodi, maybe useless
                for _alias in self.container._aliases[_type]:
                    if _alias in context.resolved:
                        _type = _alias

            # if we already have a resolver, we're done.
            if _type in context.resolved:
                return context.resolved[_type]

            reg = self.container._map.get(_type)
            if reg is None:
                continue
            resolver = reg(context)

            # add the resolver to the context, so we can find it next time we need it
            context.resolved[_type] = resolver
            return resolver

        raise CannotResolveTypeException(desired_type)

    @cached_property
    def base_type(self):
        return import_string(self.service.base) if self.service.base else self.concrete_type

    @cached_property
    def concrete_type(self) -> Type:
        _type = self.service.type or self.service.base
        if isinstance(_type, str):
            return import_string(_type)
        for candidate in _type:
            if candidate is not None and isinstance(candidate, str):
                return import_string(candidate)
        raise ValueError(f"Could not resolve concrete type for service {self.service.name}.")

    @cached_property
    def positionals(self):
        return self.service.positionals if self.service.positionals is not None else ()

    @cached_property
    def arguments(self):
        return self.service.arguments if self.service.arguments is not None else {}

    @cached_property
    def defaults(self):
        return self.service.defaults if self.service.defaults is not None else {}

    @cached_property
    def constructor(self):
        return self.service.constructor

    @cached_property
    def signature(self):
        # try to build the signature object, or return None if it is not possible. Cython objects for example does not
        # have signatures (see https://cython.readthedocs.io/en/latest/src/userguide/limitations.html#inspect-support)
        if self.constructor:
            try:
                return signature(getattr(self.concrete_type, self.constructor))
            except ValueError:
                return None
        try:
            return signature(self.concrete_type)
        except ValueError:
            return None

    def __call__(self, context: ResolutionContext):
        concrete_type = self.concrete_type

        chain = context.dynamic_chain
        chain.append(concrete_type)

        args = []
        for arg in self.positionals:
            if isinstance(arg, LazyServiceReference):
                args.append(self._get_resolver(arg.target, context))
            else:
                args.append(arg)

        kwargs = {}
        for _name, _value in self.defaults.items():
            if isinstance(_value, LazyServiceReference | LazyService):
                kwargs[_name] = _value.resolve(self._get_resolver, context)
            else:
                kwargs[_name] = _value

        if self.signature:
            for _name, _value in self.signature.parameters.items():
                if _value.annotation in self.container._map:
                    kwargs.setdefault(_name, self._get_resolver(_value.annotation, context))

        for _name, _value in self.arguments.items():
            if isinstance(_value, LazyServiceReference | LazyService):
                kwargs[_name] = _value.resolve(self._get_resolver, context)
            else:
                kwargs[_name] = _value

        return ServiceProvider(
            concrete_type,
            self.constructor,
            args=args,
            kwargs=kwargs,
            lifestyle=self.lifestyle,
        )

    def __repr__(self):
        return f"<{type(self).__name__} for {self.base_type.__name__}>"



import httpx
from hishel import ParseError
from httpx import RequestNotRead, ResponseNotRead
from rich.console import Console
from rich.syntax import Syntax

from harp.http import HttpRequestSerializer
from harp.http.serializers import HttpResponseSerializer
from harp.http.utils import parse_cache_control
from harp.utils.strings import truncate_string
from harp.utils.types import typeof
from harp_apps.http_client.events import HttpClientFilterEvent
from harp_apps.proxy.events import ProxyFilterEvent

BODY_MAX_LENGTH_TO_DISPLAY = 4096

console = Console()


async def on_proxy_request_dump(event: ProxyFilterEvent):
    serializer = HttpRequestSerializer(event.request)
    await event.request.aread()
    console.print(f"[blue]▶[/blue] Request ({typeof(event.request)})")
    console.print(
        Syntax(
            "\n".join(
                (
                    serializer.summary,
                    serializer.headers,
                    truncate_string(serializer.body.decode(), BODY_MAX_LENGTH_TO_DISPLAY),
                )
            ).strip(),
            "http",
            background_color="default",
        )
    )


async def on_proxy_response_dump(event: ProxyFilterEvent):
    serializer = HttpResponseSerializer(event.response)
    await event.response.aread()
    console.print(f"[green]◀[/green] Proxy Response ({typeof(event.response)})")
    console.print(
        Syntax(
            "\n".join(
                (
                    serializer.summary,
                    serializer.headers,
                    truncate_string(serializer.body.decode(), BODY_MAX_LENGTH_TO_DISPLAY),
                )
            ).strip(),
            "http",
            background_color="default",
        )
    )


def dump_httpx_request(request: httpx.Request) -> str:
    # Convert the request headers to a string
    headers_str = "\n".join(f"{name}: {value}" for name, value in request.headers.items())

    # Access the request body; for non-streaming bodies, you can directly use request.content
    # For streaming bodies, you might need to handle them differently depending on your use case
    try:
        body_str = request.content.decode("utf-8") if request.content else ""
    except RequestNotRead:
        body_str = ""
    body_str = truncate_string(body_str, BODY_MAX_LENGTH_TO_DISPLAY)

    # Combine method, URL, headers, and body into a formatted string
    request_dump = f"{request.method} {request.url.raw_path.decode()} HTTP/1.1\n{headers_str}\n\n{body_str}"

    return request_dump.strip()


def dump_httpx_response(response: httpx.Response) -> str:
    # Convert the response headers to a string
    headers_str = "\n".join(f"{name}: {value}" for name, value in response.headers.items())

    # Access the response body; for non-streaming bodies, you can directly use response.content
    # For streaming bodies, you might need to handle them differently depending on your use case
    try:
        body_str = response.text
    except ResponseNotRead:
        body_str = ""
    body_str = truncate_string(body_str, BODY_MAX_LENGTH_TO_DISPLAY)

    # Combine status code, headers, and body into a formatted string
    response_dump = (
        f"{response.http_version} {response.status_code} {response.reason_phrase}\n{headers_str}\n\n{body_str}"
    )

    return response_dump.strip()


async def on_remote_request_dump(event: HttpClientFilterEvent):
    console.print(f"[blue]▶▶[/blue] Remote Request ({typeof(event.request)})")
    console.print(
        Syntax(
            dump_httpx_request(event.request),
            "http",
            background_color="default",
        )
    )


async def on_remote_response_dump(event: HttpClientFilterEvent):
    console.print(f"[green]◀◀[/green] Remote Response ({typeof(event.response)})")
    console.print(
        Syntax(
            dump_httpx_response(event.response),
            "http",
            background_color="default",
        )
    )


async def on_remote_response_show_cache_control(event: HttpClientFilterEvent):
    response = event.response
    cache_control = response.headers.get_list("Cache-Control")
    if cache_control:
        try:
            parsed_cached_control = parse_cache_control(cache_control)
        except ParseError as exc:
            parsed_cached_control = exc

        console.print("[green]◀◀[/green] Cache Control")
        console.print(
            Syntax(
                f"{typeof(parsed_cached_control, short=True)}: {parsed_cached_control}",
                "http",
                background_color="default",
            )
        )



from typing import cast

from click import Command
from rich.tree import Tree

from harp import get_logger
from harp.config.examples import get_available_examples
from harp.utils.commandline import click
from harp.utils.console import console

logger = get_logger(__name__)


@click.group("examples")
def entrypoint():
    """Examples related commands."""
    pass


@click.command("list")
@click.option("--raw", is_flag=True, help="Simple output without any decoration.")
def list_command(raw=False):
    """List available examples."""

    if raw:
        for example in get_available_examples():
            click.echo(example)
    else:
        tree = Tree("List of available examples")
        for example in get_available_examples():
            tree.add(example)
        console.print(tree)


list_command = cast(Command, list_command)

entrypoint.add_command(list_command)


entrypoint = cast(Command, entrypoint)



from functools import cached_property

from harp.config import Configurable, Stateful

from .endpoint import Endpoint, EndpointSettings
from .remote import Remote, RemoteEndpoint, RemoteEndpointSettings, RemoteProbe, RemoteProbeSettings, RemoteSettings

__all__ = [
    "Endpoint",
    "EndpointSettings",
    "Proxy",
    "ProxySettings",
    "Remote",
    "RemoteEndpoint",
    "RemoteEndpointSettings",
    "RemoteProbe",
    "RemoteProbeSettings",
    "RemoteSettings",
]


class BaseProxySettings(Configurable):
    pass


class ProxySettings(BaseProxySettings):
    """
    Configuration parser for ``proxy`` settings.

    .. code-block:: yaml

        endpoints:
          # see ProxyEndpoint
          - ...

    """

    endpoints: list[EndpointSettings] = []


class Proxy(Stateful[ProxySettings]):
    @cached_property
    def endpoints(self) -> list[Endpoint]:
        return [Endpoint(settings=settings) for settings in self.settings.endpoints]



import time
from http.client import HTTPConnection
from subprocess import PIPE, Popen

import pytest


@pytest.fixture(scope="session")
def process():
    process = Popen(["make", "start"], stdout=PIPE)
    retries = 5
    while retries > 0:
        conn = HTTPConnection("localhost:4080")
        try:
            conn.request("HEAD", "/")
            response = conn.getresponse()
            if response is not None:
                yield process
                break
        except ConnectionRefusedError:
            time.sleep(1)
            retries -= 1

    if not retries:
        raise RuntimeError("Failed to start http server")
    else:
        process.terminate()
        process.wait()



import asyncio


async def check_output(*args, **kwargs):
    p = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        **kwargs,
    )
    stdout_data, stderr_data = await p.communicate()

    if p.returncode == 0:
        return stdout_data



from contextlib import asynccontextmanager

from redis.asyncio import Redis


@asynccontextmanager
async def create_redis_client() -> Redis:
    from testcontainers.redis import AsyncRedisContainer

    with AsyncRedisContainer() as redis_container:
        client = await redis_container.get_async_client()
        try:
            yield client
        finally:
            await client.aclose()



import sys
from argparse import Namespace
from contextlib import contextmanager
from os.path import basename, dirname, join
from string import Template
from tempfile import NamedTemporaryFile
from types import ModuleType

from harp.services import Container
from harp.utils.packages import get_full_qualified_name


class Stub:
    pass


class AnotherStub:
    pass


base_content = f'services: [ {{ name: "foo", type: "{get_full_qualified_name(Stub)}" }} ]'


@contextmanager
def _create_temporary_configuration_files(template: Template):
    with (
        NamedTemporaryFile(delete_on_close=False, suffix=".yaml") as base,
        NamedTemporaryFile(delete_on_close=False, suffix=".yaml") as child,
    ):
        base.write(base_content.encode())
        base.close()
        child.write(template.substitute(filename=base.name, basename=basename(base.name)).encode())
        child.close()
        yield Namespace(base=base, child=child)


def test_container_load_include():
    with _create_temporary_configuration_files(Template("services: [ !include $filename ]")) as files:
        container = Container()
        container.load(files.child.name, bind_settings={})
        provider = container.build_provider()
        assert isinstance(provider.get("foo"), Stub)


def test_container_load_include_override():
    with _create_temporary_configuration_files(
        Template(
            f'services: [ !include $filename, {{ name: "foo", override: "replace", type: "{get_full_qualified_name(AnotherStub)}" }} ]'
        )
    ) as files:
        container = Container()
        container.load(files.child.name, bind_settings={})
        provider = container.build_provider()
        assert isinstance(provider.get("foo"), AnotherStub)


def test_container_load_include_from_module():
    m = ModuleType("foo")
    sys.modules[m.__name__] = m

    with _create_temporary_configuration_files(
        Template(f'services: [ !include "$basename from {m.__name__}" ]')
    ) as files:
        m.__file__ = join(dirname(files.child.name), "__init__.py")
        m.__path__ = [dirname(files.child.name)]

        container = Container()
        container.load(files.child.name, bind_settings={})
        provider = container.build_provider()
        instance = provider.get("foo")

        assert isinstance(instance, Stub)



import httpx
import pytest
import respx
from httpx import ConnectTimeout

from harp.config.asdict import asdict
from harp.utils.testing.config import BaseConfigurableTest
from harp_apps.proxy.settings.remote import Remote, RemoteSettings


class TestRemoteSettings(BaseConfigurableTest):
    type = RemoteSettings
    expected_verbose = {
        "break_on": ["network_error", "unhandled_exception"],
        "check_after": 10.0,
        "endpoints": None,
        "min_pool_size": 1,
        "probe": None,
        "liveness": {"type": "inherit"},
    }

    def test_validate_break_on(self):
        obj = self.create(break_on={"http_5xx", "network_error"})
        assert asdict(obj) == {"break_on": ["http_5xx", "network_error"]}

    def test_validate_break_on_invalid(self):
        with pytest.raises(ValueError):
            self.create(break_on={"invalid_value"})

    def test_getitem(self):
        obj = self.create(endpoints=[{"url": "http://example.com"}])
        assert obj["http://example.com"] == obj.endpoints[0]
        assert obj["http://example.com/"] == obj.endpoints[0]
        with pytest.raises(KeyError):
            _ = obj["http://invalid.com"]

    def test_asdict(self):
        remote = self.create(endpoints=[{"url": "http://example.com"}])
        assert asdict(remote) == {"endpoints": [{"url": "http://example.com/"}]}

    def test_asdict_with_nondefault_poolsize_and_thresholds(self):
        data = {
            "endpoints": [
                {
                    "url": "http://example.com/",
                    "liveness": {
                        "type": "naive",
                        "success_threshold": 2,
                        "failure_threshold": 4,
                    },
                }
            ],
            "min_pool_size": 2,
        }
        remote = self.create(**data)
        assert asdict(remote) == data
        assert asdict(RemoteSettings(**asdict(remote))) == asdict(remote)


class TestRemoteStateful(BaseConfigurableTest):
    type = Remote
    initial = {"settings": TestRemoteSettings.initial}
    expected = {
        "current_pool": [],
        "endpoints": [],
        "settings": {
            "break_on": ["network_error", "unhandled_exception"],
            "check_after": 10.0,
            "min_pool_size": 1,
        },
    }

    expected_verbose = {
        "current_pool": [],
        "current_pool_name": "default",
        "endpoints": [],
        "probe": None,
        "settings": {
            "break_on": ["network_error", "unhandled_exception"],
            "check_after": 10.0,
            "min_pool_size": 1,
        },
    }

    def test_remote_round_robin(self):
        remote = self.create(
            settings={
                "endpoints": [
                    {"url": "http://api0.example.com/"},
                    {"url": "http://api1.example.com/"},
                ]
            }
        )

        assert remote.get_url() == "http://api0.example.com/"
        assert remote.get_url() == "http://api1.example.com/"
        assert remote.get_url() == "http://api0.example.com/"

        remote.set_down("http://api1.example.com/")
        assert remote.get_url() == "http://api0.example.com/"
        assert remote.get_url() == "http://api0.example.com/"

        remote.set_up("http://api1.example.com/")
        assert remote.get_url() == "http://api0.example.com/"
        assert remote.get_url() == "http://api1.example.com/"

    def test_remote_fallback(self):
        remote = self.create(
            settings={
                "endpoints": [
                    {"url": "http://api0.example.com/"},
                    {"url": "http://api1.example.com/"},
                    {"url": "http://fallback.example.com/", "pools": ["fallback"]},
                ],
                "min_pool_size": 2,
            }
        )

        assert remote.get_url() == "http://api0.example.com/"
        assert remote.get_url() == "http://api1.example.com/"
        assert remote.get_url() == "http://api0.example.com/"
        assert remote.get_url() == "http://api1.example.com/"

        remote.set_down("http://api1.example.com/")
        assert remote.get_url() == "http://api0.example.com/"
        assert remote.get_url() == "http://fallback.example.com/"
        assert remote.get_url() == "http://api0.example.com/"
        assert remote.get_url() == "http://fallback.example.com/"

        remote.set_up("http://api1.example.com/")
        assert remote.get_url() == "http://api0.example.com/"
        assert remote.get_url() == "http://api1.example.com/"
        assert remote.get_url() == "http://api0.example.com/"
        assert remote.get_url() == "http://api1.example.com/"

    def test_empty_pool(self):
        remote = self.create(settings={})
        with pytest.raises(IndexError):
            remote.get_url()

    def test_empty_pool_after_set_down(self):
        remote = self.create(settings={"endpoints": [{"url": "http://example.com"}]})
        assert remote.get_url() == "http://example.com/"
        remote.set_down("http://example.com")
        with pytest.raises(IndexError):
            remote.get_url()
        remote.set_up("http://example.com")
        assert remote.get_url() == "http://example.com/"

    @respx.mock
    async def test_basic_probe(self):
        remote = self.create(
            settings={
                "endpoints": [{"url": "https://example.com", "liveness": {"type": "naive", "failure_threshold": 3}}],
                "probe": {"method": "GET", "path": "/health"},
            }
        )
        healthcheck = respx.get("https://example.com/health")
        url = remote["https://example.com"]

        # initial status is "checking"
        assert url.status == 0

        # service up :)
        healthcheck.mock(return_value=httpx.Response(200))

        # first check is performed, then the status should be up, (success threshold is 1)
        await remote.check()
        assert url.status > 0

        # subsequent checks call the health endpoint, and keep the status up
        await remote.check()
        assert url.status > 0

        # service down :(
        healthcheck.mock(return_value=httpx.Response(418))

        # if the health endpoint returns a 4xx or 5xx status, we start counting failures
        await remote.check()
        assert url.status > 0
        await remote.check()
        assert url.status > 0
        await remote.check()
        assert url.status < 0

        # service up again (what a relief) :)
        healthcheck.mock(return_value=httpx.Response(200))
        await remote.check()
        assert url.status > 0

    @respx.mock
    async def test_probe_errors(self):
        remote = self.create(
            settings={
                "endpoints": [{"url": "https://example.com", "liveness": {"type": "naive", "failure_threshold": 3}}],
                "probe": {"method": "GET", "path": "/health"},
            }
        )

        healthcheck = respx.get("https://example.com/health")
        url = remote["https://example.com"]

        # initial status is "checking"
        assert url.status == 0

        # service in error ...
        healthcheck.mock(side_effect=TimeoutError)

        await remote.check()
        await remote.check()
        assert url.status == 0  # still checking, under failure threshold
        assert url.failure_reasons == {"PROBE_TIMEOUT_ERROR"}

        await remote.check()
        assert url.status < 0  # threshold reached
        assert url.failure_reasons == {"PROBE_TIMEOUT_ERROR"}

        healthcheck.mock(return_value=httpx.Response(200))
        await remote.check()
        assert url.status > 0
        assert url.failure_reasons is None

        healthcheck.mock(return_value=httpx.Response(418))
        await remote.check()
        await remote.check()
        assert url.status > 0  # still up, under threshold
        assert url.failure_reasons == {"PROBE_HTTP_418"}

        await remote.check()
        assert url.status < 0  # threshold reached
        assert url.failure_reasons == {"PROBE_HTTP_418"}

        healthcheck.mock(return_value=httpx.Response(200))
        await remote.check()
        assert url.status > 0
        assert url.failure_reasons is None

    @respx.mock
    async def test_probe_timeout(self):
        healthcheck = respx.get("https://example.com/health")
        healthcheck.mock(side_effect=ConnectTimeout)

        remote = self.create(
            settings={
                "endpoints": [{"url": "https://example.com", "liveness": {"type": "naive", "failure_threshold": 3}}],
                "probe": {"method": "GET", "path": "/health", "timeout": 0.1},
            }
        )

        url = remote["https://example.com"]

        await remote.check()
        assert url.status == 0
        assert url.failure_reasons == {"PROBE_CONNECT_TIMEOUT"}



import pytest
from decimal import Decimal

from harp.http import HttpRequest, HttpResponse
from harp_apps.proxy.events import ProxyFilterEvent


def valid_response_script(context):
    context["response"] = HttpResponse("Ok.", status=200)


def invalid_response_script(context):
    context["response"] = Decimal(200)


def none_response_script(context):
    context["response"] = None


def test_execution():
    event = ProxyFilterEvent("api", request=HttpRequest())

    event.execute_script(valid_response_script)
    assert event.response.status == 200

    with pytest.raises(ValueError):
        event.execute_script(invalid_response_script)
    assert event.response.status == 200

    event.execute_script(none_response_script)
    assert event.response is None



from typing import TYPE_CHECKING

from asgiref.typing import ASGISendCallable
from multidict import CIMultiDictProxy

from harp.utils.bytes import ensure_bytes

if TYPE_CHECKING:
    from harp.http import HttpResponse


class HttpResponseAsgiBridge:  # todo protocol HttpResponseBridge
    """Implements the ability of sending our HttpResponse object over the asgi protocol. This is still early impl. and
    it will need to support streaming responses in the future."""

    def __init__(self, response: "HttpResponse", send: ASGISendCallable):
        self.response = response
        self.asgi_send = send

    async def send(self):
        # set the headers as read only
        self.response._headers = CIMultiDictProxy(self.response._headers)

        # prepare headers for asgi
        headers = tuple((ensure_bytes(k), ensure_bytes(v)) for k, v in self.response.headers.items())

        # start the response
        await self.asgi_send(
            {
                "type": "http.response.start",
                "status": self.response.status,
                "headers": headers,
            }
        )

        # send the body
        await self.asgi_send(
            {
                "type": "http.response.body",
                "body": ensure_bytes(self.response.body),
            }
        )



"""Storage Application"""

from functools import partial
from os.path import dirname
from pathlib import Path

from sqlalchemy.ext.asyncio import AsyncEngine

from harp import get_logger
from harp.config import Application
from harp.config.events import OnBindEvent, OnBoundEvent, OnShutdownEvent
from harp_apps.storage.models import Base
from harp_apps.storage.settings import StorageSettings
from harp_apps.storage.types import IStorage
from harp_apps.storage.worker import StorageAsyncWorkerQueue

logger = get_logger(__name__)


async def _run_migrations(engine: AsyncEngine):
    if engine.dialect.name == "sqlite" and engine.url.database == ":memory:":  # pragma: no cover
        # in memory sqlite won't be able to reconnect to the same instance, so we create the tables directly
        async with engine.connect() as conn:
            await conn.run_sync(Base.metadata.create_all)
            await conn.commit()

    else:
        # todo refactor ? see harp_apps.storage.utils.testing.mixins.StorageTestFixtureMixin
        from alembic import command

        from harp_apps.storage.utils.migrations import create_alembic_config, do_migrate

        alembic_cfg = create_alembic_config(engine.url.render_as_string(hide_password=False))
        migrator = partial(command.upgrade, alembic_cfg, "head")
        await do_migrate(engine, migrator=migrator)


async def on_bind(event: OnBindEvent):
    # load service definitions, bound to our settings
    event.container.load(Path(dirname(__file__)) / "services.yml", bind_settings=event.settings["storage"])


async def on_bound(event: OnBoundEvent):
    settings = event.provider.get(StorageSettings)
    if settings.migrate:
        engine = event.provider.get(AsyncEngine)
        await _run_migrations(engine)

    storage = event.provider.get(IStorage)
    await storage.initialize()
    await storage.ready()
    worker = event.provider.get(StorageAsyncWorkerQueue)
    if event.dispatcher:
        worker.register_events(event.dispatcher)


async def on_shutdown(event: OnShutdownEvent):
    worker = event.provider.get(StorageAsyncWorkerQueue)
    await worker.wait_until_empty()
    await event.provider.get(IStorage).finalize()


application = Application(
    settings_type=StorageSettings,
    on_bind=on_bind,
    on_bound=on_bound,
    on_shutdown=on_shutdown,
)



from copy import deepcopy

from harp.config import ConfigurationBuilder
from harp.config.asdict import asdict
from harp.config.builders.system import System

from ..controllers.blobs import BlobsController
from ..settings import DashboardSettings


class BaseSystemTest:
    applications = []
    settings = {}

    async def create_system(self) -> System:
        return await ConfigurationBuilder(
            {
                "applications": self.applications,
                **deepcopy(self.settings),
            },
            use_default_applications=False,
        ).abuild_system()


class TestBlobsControllerService(BaseSystemTest):
    applications = ["http_client", "storage", "harp_apps.dashboard"]
    settings = {"dashboard": asdict(DashboardSettings(enable_ui=False))}

    async def test_get_instances(self):
        system = await self.create_system()

        # we can get an api controller for blobs-related stuff
        controller = system.provider.get("dashboard.controller.blobs")
        assert isinstance(controller, BlobsController)

        # the router is a singleton, for this application
        router = system.provider.get("dashboard.router")
        assert router is controller.router



from functools import cached_property
from inspect import BoundArguments, Parameter, Signature, signature
from typing import Optional, Type, TypeVar, Union

from rodi import (
    ActivationScope,
    ArgsTypeProvider,
    FactoryTypeProvider,
    InstanceProvider,
    ScopedArgsTypeProvider,
    ScopedFactoryTypeProvider,
    ScopedTypeProvider,
    ServiceLifeStyle,
    SingletonFactoryTypeProvider,
    SingletonTypeProvider,
    TypeProvider,
)

T = TypeVar("T")


def filter_kwargs_based_on_signature(kwargs, signature: Signature):
    """Filter out kwargs that are not in the signature or that are not allowed as keyword arguments."""
    return {
        k: v
        for k, v in kwargs.items()
        if k in signature.parameters
        and signature.parameters[k].kind in (Parameter.KEYWORD_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
    }


class ServiceProvider:
    def __init__(self, _type, _constructor=None, *, args, kwargs, lifestyle):
        self._type = _type
        self._constructor = _constructor
        self._args = args
        self._kwargs = kwargs
        self._lifestyle = lifestyle

        if self._lifestyle == ServiceLifeStyle.SINGLETON:
            self._instance = None

    @cached_property
    def constructor(self):
        return self._type if not self._constructor else getattr(self._type, self._constructor)

    @property
    def signature(self):
        try:
            sig = signature(self.constructor)
        except ValueError:
            # for objects that cannot give their signatures, we try to forge one. Of course, this will be limited, but
            # we need to somehow support it because of cython objects, for example.
            sig = Signature(
                [
                    *(Parameter(str(i), Parameter.POSITIONAL_ONLY) for i in range(len(self._args))),
                    *(Parameter(k, Parameter.KEYWORD_ONLY) for k in self._kwargs),
                ]
            )
        return sig

    def _resolve_arguments(self, scope: ActivationScope, parent_type) -> BoundArguments:
        """Create a bound argument object after resolving all the arguments (aka transforming "provider" type values
        into their actual alive counterpart."""

        def _resolve(arg):
            if isinstance(arg, PROVIDER_TYPES):
                return arg(scope, parent_type)
            return arg

        resolved_args = (_resolve(v) for v in self._args)
        resolved_kwargs = {k: _resolve(v) for k, v in {**self._kwargs}.items()}
        try:
            return self.signature.bind(*resolved_args, **resolved_kwargs)
        except TypeError as exc:
            raise TypeError(f"Error resolving arguments for {self._type.__name__}: {exc}") from exc

    def _create_instance(self, scope: ActivationScope, parent_type):
        """Create an instance of the service each time it is called."""
        arguments = self._resolve_arguments(scope, parent_type=parent_type)
        return self.constructor(
            *arguments.args,
            **arguments.kwargs,
        )

    def __call__(self, scope: ActivationScope, parent_type: Optional[Union[Type[T], str]] = None):
        """Resolves this provider into a service instance, creating it if necessary (will depend on service's life
        style)."""
        parent_type = parent_type or self._type

        # singleton lifestyle will get instanciated only once (by provider, two different providers for the same service
        # would create 2 instances-.
        if self._lifestyle == ServiceLifeStyle.SINGLETON:
            if not self._instance:
                self._instance = self._create_instance(scope, parent_type)
            return self._instance

        # scoped lifestyle will get instanciated only once per scope (for example, a web request)
        if self._lifestyle == ServiceLifeStyle.SCOPED:
            if self._type not in scope.scoped_services:
                scope.scoped_services[self._type] = self._create_instance(scope, parent_type)
            return scope.scoped_services[self._type]

        # default / transient lifestyle will get instanciated each time it is called
        return self._create_instance(scope, parent_type)

    def bind(self, **kwargs):
        """Add additional arguments to the service provider.
        We check that arguments are in the signature of the service type before adding them.
        """
        sig = self.signature
        filtered_kwargs = filter_kwargs_based_on_signature(kwargs, sig)
        self._kwargs.update(filtered_kwargs)


PROVIDER_TYPES = (
    ServiceProvider,
    InstanceProvider,
    TypeProvider,
    ScopedTypeProvider,
    ArgsTypeProvider,
    FactoryTypeProvider,
    SingletonFactoryTypeProvider,
    ScopedFactoryTypeProvider,
    ScopedArgsTypeProvider,
    SingletonTypeProvider,
)



from unittest.mock import Mock

from harp.utils.testing.communicators import ASGICommunicator

from .._base import BaseRulesFlowTest


class TestProxyRulesFlow(BaseRulesFlowTest):
    applications = ["http_client", "proxy", "rules"]

    async def test_proxy_flow(self, httpbin):
        mock = Mock()
        system = await self.create_system(
            {"proxy": {"endpoints": [{"name": "httpbin", "port": 80, "url": httpbin}]}},
            mock=mock,
        )

        client = ASGICommunicator(system.asgi_app)
        await client.asgi_lifespan_startup()

        await client.http_get("/")

        assert mock.call_count == 4

        assert mock.call_args_list[0].args[0]["rule"] == "on_request"
        assert mock.call_args_list[0].args[0]["response"] is None
        request_transaction = mock.call_args_list[0].args[0]["transaction"]
        assert request_transaction is not None

        assert mock.call_args_list[1].args[0]["rule"] == "on_remote_request"
        assert mock.call_args_list[1].args[0]["response"] is None

        assert mock.call_args_list[2].args[0]["rule"] == "on_remote_response"
        assert mock.call_args_list[2].args[0]["response"].status_code == 200

        assert mock.call_args_list[3].args[0]["rule"] == "on_response"
        assert mock.call_args_list[3].args[0]["response"].status == 200
        response_transaction = mock.call_args_list[3].args[0]["transaction"]
        assert response_transaction is not None
        assert request_transaction.id == response_transaction.id



from datetime import datetime
from typing import TypedDict


class TransactionsGroupedByTimeBucket(TypedDict):
    datetime: datetime
    count: int
    errors: int
    cached: int
    meanDuration: float
    meanTpdex: float



import random
import string
from unittest.mock import AsyncMock

import pytest
from sqlalchemy.exc import DatabaseError, ProgrammingError

from harp.utils.testing.databases import parametrize_with_database_urls
from harp_apps.storage.optionals.pg_trgm import PgTrgmOptional
from harp_apps.storage.utils.sql import run_sql
from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin
from harp_apps.storage.utils.testing.sql import run_cli_migrate_command


async def get_indexes(engine, table_name):
    return [
        row[0]
        for row in (
            await run_sql(
                engine,
                f"SELECT indexname FROM pg_indexes WHERE tablename = '{table_name}';",
            )
        ).fetchall()
    ]


def generate_random_db_name(length=5):
    characters = string.ascii_lowercase + string.digits
    return "test_" + "".join(random.choice(characters) for _ in range(length))


@parametrize_with_database_urls("postgresql")
async def test_pg_trgm(database_url):
    opt = PgTrgmOptional(database_url)

    # cannot execute because it needs the database to contain something first.
    with pytest.raises(ProgrammingError):
        await opt.install()

    # run the migrations
    result = await run_cli_migrate_command(database_url)
    assert result.exit_code == 0

    # we should not have gin indexes yet
    assert "messages_summary_gin" not in await get_indexes(opt.engine, "messages")
    assert "transactions_endpoint_gin" not in await get_indexes(opt.engine, "transactions")

    # now we can actually install our optimized indexes
    await opt.install()

    # extension and indexes should exist now
    assert await opt._is_pg_trgm_extension_installed()
    assert "messages_summary_gin" in await get_indexes(opt.engine, "messages")
    assert "transactions_endpoint_gin" in await get_indexes(opt.engine, "transactions")

    # let's uninstall
    await opt.uninstall()

    # we should not have gin indexes anymore, but the extension should still be there
    assert await opt._is_pg_trgm_extension_installed()
    assert "messages_summary_gin" not in await get_indexes(opt.engine, "messages")
    assert "transactions_endpoint_gin" not in await get_indexes(opt.engine, "transactions")


@parametrize_with_database_urls("postgresql")
async def test_pg_trgm_on_database_without_create_extension_privilege(database_url):
    opt = PgTrgmOptional(database_url)

    # simulate a database without the privilege to create extensions
    opt._is_pg_trgm_extension_installed = AsyncMock(return_value=False)
    opt._install_pg_trgm_extension_if_possible = AsyncMock(side_effect=DatabaseError("nope.", None, Exception()))
    assert not await opt._is_pg_trgm_extension_installed()

    # run the migrations
    result = await run_cli_migrate_command(database_url)
    assert result.exit_code == 0

    # we should not have gin indexes ...
    assert "messages_summary_gin" not in await get_indexes(opt.engine, "messages")
    assert "transactions_endpoint_gin" not in await get_indexes(opt.engine, "transactions")

    await opt.install()

    # ... but neither after upgrade
    assert "messages_summary_gin" not in await get_indexes(opt.engine, "messages")
    assert "transactions_endpoint_gin" not in await get_indexes(opt.engine, "transactions")


@parametrize_with_database_urls("mysql", "sqlite")
async def test_pg_trgm_not_available_for_non_postgres_databases(database_url):
    opt = PgTrgmOptional(database_url)

    # cannot install on not supported dialects
    with pytest.raises(RuntimeError):
        await opt.install()


class TestOptimizedQueries(StorageTestFixtureMixin):
    @parametrize_with_database_urls("postgresql")
    async def test_optimized_queries(self, sql_storage):
        # store raw sql queries in `storage.sql_queries`, please
        sql_storage.install_debugging_instrumentation()

        opt = PgTrgmOptional(sql_storage.engine.url)
        try:
            await opt.install()

            result = await sql_storage.get_transaction_list(username="anonymous", with_messages=True, text_search="bar")

            # count + select
            assert len(sql_storage.sql_queries) == 2

            for _ in range(100):
                await self.create_transaction(
                    sql_storage,
                    endpoint="/api/transactions/foo/bar/baz",
                    tags={"foo": "bar"},
                )

            result = await sql_storage.get_transaction_list(username="anonymous", with_messages=True, text_search="bar")

            assert len(result) == 40
            # first queries, then count + select + flags,tags,messages
            assert len(sql_storage.sql_queries) == 2 + 2 + 3

        finally:
            await opt.engine.dispose()



import pytest
from httpx import ByteStream

from harp.http import HttpRequest, HttpRequestSerializer


class BaseHttpRequestTest:
    """
    Base class for testing HTTP requests, without focusing on a specific underlying protocol bridge implementation.
    """

    RequestType = HttpRequest

    def create_request(self, **kwargs) -> HttpRequest:
        return self.RequestType(**kwargs)


class TestHttpRequestPath(BaseHttpRequestTest):
    """
    Test path-related stuff on HTTP requests.
    """

    def test_default(self):
        request = self.create_request()
        assert request.path == "/"

    def test_simple(self):
        request = self.create_request(path="/foo/bar")
        assert request.path == "/foo/bar"


class TestHttpRequestQuery(BaseHttpRequestTest):
    """
    Test query string-related stuff on HTTP requests.
    """

    def test_simple(self):
        request = self.create_request(
            query="foo=bar&baz=qux",
        )

        assert list(sorted(request.query.items())) == [("baz", "qux"), ("foo", "bar")]

    def test_multi_keys(self):
        request = self.create_request(
            query="foo=bar&baz=qux&foo=more&foo=stuff&john=doe",
        )

        assert list(sorted(request.query.items())) == [
            ("baz", "qux"),
            ("foo", "bar"),
            ("foo", "more"),
            ("foo", "stuff"),
            ("john", "doe"),
        ]

    def test_multi_values(self):
        request = self.create_request(
            query="foo=bar&foo=bar&foo=bar",
        )

        assert list(sorted(request.query.items())) == [
            ("foo", "bar"),
            ("foo", "bar"),
            ("foo", "bar"),
        ]


class TestHttpRequestBasicAuthentication(BaseHttpRequestTest):
    """
    Test basic authentication-related stuff on HTTP requests.

    (cf https://developer.mozilla.org/fr/docs/Web/HTTP/Headers/Authorization)
    """

    def test_default_is_unset(self):
        request = self.create_request()
        assert request.basic_auth is None

    def test_simple(self):
        request = self.create_request(headers={"authorization": "Basic dXNlcjpwYXNz"})
        assert request.basic_auth == ("user", "pass")


class TestHttpRequestHeaders(BaseHttpRequestTest):
    """
    Test headers-related stuff on HTTP requests (except cookie which will get their own home).
    """

    def test_headers(self):
        request = self.create_request(
            headers={
                "host": "localhost:4080",
                "connection": "keep-alive",
                "origin": "http://localhost:4080",
                "accept": "*/*",
                "referer": "http://localhost:4080/",
                "accept-encoding": "gzip, deflate, br",
                "accept-language": "fr-FR,fr;q=0.9,en;q=0.8,fr-CA;q=0.7",
            }
        )

        assert request.headers == {
            "host": "localhost:4080",
            "connection": "keep-alive",
            "origin": "http://localhost:4080",
            "accept": "*/*",
            "referer": "http://localhost:4080/",
            "accept-encoding": "gzip, deflate, br",
            "accept-language": "fr-FR,fr;q=0.9,en;q=0.8,fr-CA;q=0.7",
        }

        # TODO move into serializer tests
        serializer = HttpRequestSerializer(request)
        assert serializer.headers == (
            "host: localhost:4080\n"
            "connection: keep-alive\n"
            "origin: http://localhost:4080\n"
            "accept: */*\n"
            "referer: http://localhost:4080/\n"
            "accept-encoding: gzip, deflate, br\n"
            "accept-language: fr-FR,fr;q=0.9,en;q=0.8,fr-CA;q=0.7"
        )


class TestHttpRequestCookies(BaseHttpRequestTest):
    """
    Test cookie-related stuff on HTTP requests.
    """

    def test_cookies_empty(self):
        request = self.create_request()
        assert request.cookies == {}

    def test_cookies_basics(self):
        request = self.create_request(headers={"cookie": "name=value; name2=value2; name3=value3"})
        assert request.cookies == {
            "name": "value",
            "name2": "value2",
            "name3": "value3",
        }

    def test_cookies_more(self):
        request = self.create_request(
            headers={
                "cookie": "bab_locale=fr; bab_original=fr; bab_block=1690; _fbp=fb.1.1687.19100; "
                "_ga_6F3R65C=GS1.1.1452.9.0.1732.0.0.0; _ga_D73X6=GS1.1.1762.30.1.1796.0.0.0; "
                "harp=421 balloons flying; _ga_X8BC7QX9TE=GS1.1.7712.22.1.1753.0.0.0; "
                "_ga=GA1.1.7767.1438; _ga_PPPB3LT24D=GS1.1.1784.84.0.1784.0.0.0",
            }
        )

        assert request.cookies == {
            "_fbp": "fb.1.1687.19100",
            "_ga": "GA1.1.7767.1438",
            "_ga_6F3R65C": "GS1.1.1452.9.0.1732.0.0.0",
            "_ga_D73X6": "GS1.1.1762.30.1.1796.0.0.0",
            "_ga_PPPB3LT24D": "GS1.1.1784.84.0.1784.0.0.0",
            "_ga_X8BC7QX9TE": "GS1.1.7712.22.1.1753.0.0.0",
            "bab_block": "1690",
            "bab_locale": "fr",
            "bab_original": "fr",
            "harp": "421 balloons flying",
        }


class TestHttpRequestBody(BaseHttpRequestTest):
    async def test_body_must_be_read_first(self):
        request = self.create_request()
        with pytest.raises(RuntimeError):
            assert request.body == b""

    async def test_body_empty(self):
        request = self.create_request()
        await request.aread()
        assert request.body == b""

    async def test_body_one_chunk(self):
        request = self.create_request(body=b"foobar")
        await request.aread()
        assert request.body == b"foobar"

    async def test_body_many_chunks(self):
        request = self.create_request(body=[b"foo", b"bar", b"baz"])
        await request.aread()
        assert request.body == b"foobarbaz"

    async def test_body_can_be_read_more_than_once(self):
        request = self.create_request(body=[b"foo", b"bar", b"baz"])
        await request.aread()
        await request.aread()
        await request.aread()
        assert request.body == b"foobarbaz"

    async def test_stream_can_be_accessed_before_reading_body(self):
        request = self.create_request(body=[b"foo", b"bar", b"baz"])
        assert [chunk async for chunk in request.stream] == [b"foo", b"bar", b"baz"]
        request.stream = ByteStream(b"foobarbaz")
        await request.aread()
        assert request.body == b"foobarbaz"

    async def test_stream_can_be_accessed_after_reading_body(self):
        request = self.create_request(body=[b"foo", b"bar", b"baz"])
        await request.aread()
        assert [chunk async for chunk in request.stream] == [b"foobarbaz"]
        assert request.body == b"foobarbaz"
        assert isinstance(request.body, bytes)
        assert isinstance(request.stream, ByteStream)



import http

import pytest

from harp.http.utils import HTTP_METHODS
from harp.typing import Maybe, NotSet


def parametrize_with_http_status_codes(include=None):
    statuses = [status.value for status in http.HTTPStatus]
    if include is not None:
        statuses = [status for status in statuses if status // 100 in include]
    return pytest.mark.parametrize("status_code", statuses)


def _filter_methods_by_attribute(attribute, value):
    return {name for name, method in HTTP_METHODS.items() if getattr(method, attribute) is value}


def _update_methods_set(methods, include, attribute, value):
    if include is True:
        methods |= _filter_methods_by_attribute(attribute, value)
    elif include is False:
        methods -= _filter_methods_by_attribute(attribute, value)


def parametrize_with_http_methods(
    *,
    include_safe=NotSet,
    include_unsafe=NotSet,
    include_idempotent=NotSet,
    include_non_idempotent=NotSet,
    include_standard=NotSet,
    include_non_standard=NotSet,
    include_having_request_body=NotSet,
    include_maybe_having_request_body=NotSet,
    include_not_having_request_body=NotSet,
    include_having_response_body=NotSet,
    include_maybe_having_response_body=NotSet,
    include_not_having_response_body=NotSet,
    exclude=(),
):
    """
    Parametrize a test with HTTP methods based on various inclusion and exclusion criteria.

    Parameters:
    - include_safe (bool or NotSet): Include safe methods (e.g., GET, HEAD).
    - include_unsafe (bool or NotSet): Include unsafe methods (e.g., POST, PUT).
    - include_idempotent (bool or NotSet): Include idempotent methods (e.g., GET, PUT).
    - include_non_idempotent (bool or NotSet): Include non-idempotent methods (e.g., POST).
    - include_standard (bool or NotSet): Include standard HTTP methods.
    - include_non_standard (bool or NotSet): Include non-standard HTTP methods.
    - include_having_request_body (bool or NotSet): Include methods that have a request body.
    - include_maybe_having_request_body (bool or NotSet): Include methods that may have a request body.
    - include_not_having_request_body (bool or NotSet): Include methods that do not have a request body.
    - include_having_response_body (bool or NotSet): Include methods that have a response body.
    - include_maybe_having_response_body (bool or NotSet): Include methods that may have a response body.
    - include_not_having_response_body (bool or NotSet): Include methods that do not have a response body.
    - exclude (tuple): Methods to exclude from the parametrization.

    Returns:
    - pytest.mark.parametrize: A pytest parametrize marker with the selected HTTP methods.

    """
    if all(
        param is NotSet
        for param in [
            include_safe,
            include_unsafe,
            include_idempotent,
            include_non_idempotent,
            include_standard,
            include_non_standard,
            include_having_request_body,
            include_maybe_having_request_body,
            include_not_having_request_body,
            include_having_response_body,
            include_maybe_having_response_body,
            include_not_having_response_body,
        ]
    ):
        include_standard = True

    methods = set()
    _update_methods_set(methods, include_safe, "safe", True)
    _update_methods_set(methods, include_unsafe, "safe", False)
    _update_methods_set(methods, include_idempotent, "idempotent", True)
    _update_methods_set(methods, include_non_idempotent, "idempotent", False)
    if include_standard is True:
        methods |= set(HTTP_METHODS.keys())
    if include_non_standard is True:
        methods |= {"BREW", "REMIX"}
    _update_methods_set(methods, include_having_request_body, "request_body", True)
    _update_methods_set(methods, include_not_having_request_body, "request_body", False)
    _update_methods_set(methods, include_maybe_having_request_body, "request_body", Maybe)
    _update_methods_set(methods, include_having_response_body, "response_body", True)
    _update_methods_set(methods, include_not_having_response_body, "response_body", False)
    _update_methods_set(methods, include_maybe_having_response_body, "response_body", Maybe)

    methods -= set(exclude)

    return pytest.mark.parametrize("method", sorted(methods))



from asgi_prometheus import PrometheusMiddleware

from harp import get_logger
from harp.config import Application, OnReadyEvent
from harp.controllers import ProxyControllerResolver

from .decorators import decorate_http_client_send, decorate_proxy_controller
from .settings import MetricsSettings

_metrics_url = "/.prometheus/metrics"

logger = get_logger(__name__)


async def on_ready(event: OnReadyEvent):
    # decorate the asgi app with outer metrics
    event.asgi_app = PrometheusMiddleware(event.asgi_app, metrics_url=_metrics_url, group_paths=["/"])
    event.asgi_app.scopes = ("http",)

    # decorate all controllers with metrics
    resolver = event.provider.get(ProxyControllerResolver)
    for i in resolver:
        resolver[i] = decorate_proxy_controller(
            resolver[i],
            name=resolver.get_controller_name_by_index(i),
            port=resolver.get_controller_port_by_index(i),
        )

    # decorate the http client with metrics
    http_client = event.provider.get("http_client")
    http_client.send = decorate_http_client_send(http_client.send)

    logger.info(f"🌎 Prometheus enabled, metrics under {_metrics_url}.")


application = Application(
    dependencies=["proxy"],
    on_ready=on_ready,
    settings_type=MetricsSettings,
)



import logging
import sys
from typing import cast

from asyncpg import PostgresError

from harp import get_logger
from harp.config.builders.system import System

logger = get_logger(__name__)


class HypercornAdapter:
    def __init__(self, system: System):
        self.system = system

    def _create_hypercorn_config(self, binds):
        """
        Creates a hypercorn config object.

        :return: hypercorn config object
        """
        from hypercorn.config import Config

        config = Config()
        config.bind = [*map(str, binds)]
        config.accesslog = logging.getLogger("hypercorn.access")
        config.errorlog = logging.getLogger("hypercorn.error")
        return config

    async def serve(self):
        """
        Creates and serves the proxy using hypercorn.
        """
        from hypercorn.asyncio import serve
        from hypercorn.typing import Framework
        from hypercorn.utils import LifespanFailureError

        hypercorn_config = self._create_hypercorn_config(self.system.binds)
        logger.debug(f"🌎 {type(self).__name__}::serve({', '.join(hypercorn_config.bind)})")

        try:
            return await serve(cast(Framework, self.system.asgi_app), hypercorn_config, mode="asgi")
        except LifespanFailureError as exc:
            logger.exception(f"Server initiliation failed: {repr(exc.__cause__)}", exc_info=exc.__cause__)
            if isinstance(exc.__cause__, PostgresError):
                logger.error("Could not connect to underlying postgres storage backend, check your config.")
            sys.exit(-1)



from harp.config.asdict import asdict
from harp.utils.testing.config import BaseConfigurableTest
from harp_apps.proxy.settings import EndpointSettings
from harp_apps.proxy.settings.endpoint import Endpoint

base_settings = {
    "name": "my-endpoint",
    "port": 8080,
}
remote_settings = {"endpoints": [{"url": "http://example.com/"}]}


class TestEndpointSettings(BaseConfigurableTest):
    type = EndpointSettings
    initial = {**base_settings}
    expected = {**initial}
    expected_verbose = {
        **expected,
        "description": None,
        "remote": None,
        "controller": {
            "type": "harp_apps.proxy.controllers.HttpProxyController",
            "dispatcher": {"type": "IAsyncEventDispatcher"},
        },
    }

    def test_old_url_syntax(self):
        obj = self.create(url="http://my-endpoint:8080")
        assert asdict(obj) == {
            **self.expected,
            "remote": {"endpoints": [{"url": "http://my-endpoint:8080/"}]},
        }


class TestEndpointSettingsWithRemote(BaseConfigurableTest):
    type = EndpointSettings
    initial = {
        **base_settings,
        "remote": remote_settings,
    }
    expected = {**initial}
    expected_verbose = {
        **expected,
        "description": None,
        "remote": {
            "break_on": ["network_error", "unhandled_exception"],
            "check_after": 10.0,
            "endpoints": [
                {
                    "liveness": {"type": "inherit"},
                    "pools": ["default"],
                    "url": "http://example.com/",
                }
            ],
            "min_pool_size": 1,
            "probe": None,
            "liveness": {"type": "inherit"},
        },
        "controller": {
            "type": "harp_apps.proxy.controllers.HttpProxyController",
            "dispatcher": {"type": "IAsyncEventDispatcher"},
        },
    }


class TestEndpointStateful(BaseConfigurableTest):
    type = Endpoint
    initial = {
        "settings": TestEndpointSettings.initial,
    }
    expected = {
        "settings": {
            **TestEndpointSettings.expected,
            "description": None,
        },
    }
    expected_verbose = {
        **expected,
        "remote": None,
    }

    def test_repr(self):
        obj = self.create()
        assert repr(obj) == "Endpoint(remote=None)"


class TestEndpointStatefulWithRemote(BaseConfigurableTest):
    type = Endpoint
    initial = {
        "settings": TestEndpointSettingsWithRemote.initial,
    }
    expected = {
        "remote": {
            "current_pool": ["http://example.com/"],
            "endpoints": [
                {
                    "settings": {
                        "liveness": {"type": "inherit"},
                        "url": "http://example.com/",
                        "pools": ["default"],
                    },
                }
            ],
            "settings": {
                "break_on": ["network_error", "unhandled_exception"],
                "check_after": 10.0,
                "min_pool_size": 1,
            },
        },
        "settings": {
            **TestEndpointSettings.expected,
            "description": None,
        },
    }

    expected_verbose = {
        "remote": {
            "current_pool": ["http://example.com/"],
            "current_pool_name": "default",
            "endpoints": [
                {
                    "failure_reasons": None,
                    "settings": {"liveness": {"type": "inherit"}, "pools": ["default"], "url": "http://example.com/"},
                    "status": 0,
                }
            ],
            "probe": None,
            "settings": {"break_on": ["network_error", "unhandled_exception"], "check_after": 10.0, "min_pool_size": 1},
        },
        "settings": expected["settings"],
    }



from harp.asgi.events import ViewEvent
from harp.http import HttpResponse


class html(str):
    content_type = "text/html"


async def on_string_response(event: ViewEvent):
    if isinstance(event.value, str):
        content_type = getattr(event.value, "content_type", "text/plain")
        event.set_response(
            HttpResponse(event.value, content_type=content_type),
        )



import os


def get_adjusted_timeout(base_timeout: float = 10.0) -> float:
    """Adjust timeout based on system load.

    Args:
        base_timeout: Base timeout in seconds.

    Returns:
        Adjusted timeout value.
    """
    try:
        # Get system load average (1, 5, 15 minutes)
        load_avg = os.getloadavg()[0]
        cpu_count = os.cpu_count() or 1

        # Calculate load factor (load per CPU)
        load_factor = load_avg / cpu_count

        # Adjust timeout based on load
        if load_factor < 1.5:
            return base_timeout
        elif load_factor < 3.0:
            return base_timeout * 1.5
        else:
            return min(base_timeout * 2.0, 30.0)
    except AttributeError:
        # Windows doesn't have getloadavg
        return base_timeout * 1.5



from harp import get_logger
from harp.config.applications import Application
from harp.config.events import OnBindEvent, OnBoundEvent, OnShutdownEvent

from .settings import RulesSettings
from .subscribers import RulesSubscriber

logger = get_logger(__name__)

RULES_SUBSCRIBER = "rules.subscriber"


async def on_bind(event: OnBindEvent):
    settings = event.settings.get("rules")
    logger.warning("📦 Rules: found %d rules.", len(settings.ruleset))


async def on_bound(event: OnBoundEvent):
    settings = event.provider.get(RulesSettings)
    subscriber = RulesSubscriber(settings.ruleset)
    subscriber.subscribe(event.dispatcher)
    event.provider.set(RULES_SUBSCRIBER, subscriber)


async def on_shutdown(event: OnShutdownEvent):
    subscriber = event.provider.get(RULES_SUBSCRIBER)
    subscriber.unsubscribe(event.dispatcher)


application = Application(
    settings_type=RulesSettings,
    on_bind=on_bind,
    on_bound=on_bound,
    on_shutdown=on_shutdown,
)



import asyncio
import re
from contextlib import asynccontextmanager
from datetime import UTC, datetime, timedelta
from operator import itemgetter
from typing import Iterable, Optional, override

from sqlalchemy import and_, bindparam, case, delete, func, literal, literal_column, null, or_, select, text
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker
from sqlalchemy.sql.functions import count

from harp import get_logger
from harp.models import Results, Transaction
from harp.settings import PAGE_SIZE
from harp.utils.background import AsyncWorkerQueue
from harp.utils.dates import ensure_datetime
from harp_apps.storage.constants import TimeBucket
from harp_apps.storage.models import (
    FLAGS_BY_NAME,
    Base,
    BlobsRepository,
    FlagsRepository,
)
from harp_apps.storage.models import Message as SqlMessage
from harp_apps.storage.models import (
    MessagesRepository,
    MetricsRepository,
    MetricValuesRepository,
    TagsRepository,
    TagValuesRepository,
)
from harp_apps.storage.models import Transaction as SqlTransaction
from harp_apps.storage.models import (
    TransactionsRepository,
)
from harp_apps.storage.models import User as SqlUser
from harp_apps.storage.models import UserFlag as SqlUserFlag
from harp_apps.storage.models import (
    UsersRepository,
)
from harp_apps.storage.settings import StorageSettings
from harp_apps.storage.types import IBlobStorage, IStorage, TransactionsGroupedByTimeBucket
from harp_apps.storage.utils.dates import TruncDatetime

logger = get_logger(__name__)

_FILTER_COLUMN_NAMES = {
    "method": "x_method",
    "status": "x_status_class",
    "tpdex": "tpdex",
}


def _numerical_filter_query(query, name: str, values: dict[str, float]):
    if values:
        column_name = _FILTER_COLUMN_NAMES.get(name, name)
        column = getattr(SqlTransaction, column_name)

        min_val = values.get("min")
        max_val = values.get("max")

        if min_val is not None and max_val is not None:
            query = query.filter(column.between(min_val, max_val))
        elif min_val is not None:
            query = query.filter(column >= min_val)
        elif max_val is not None:
            query = query.filter(column <= max_val)

    return query


def _filter_query(query, name, values):
    if values:
        query = query.filter(
            getattr(
                SqlTransaction,
                _FILTER_COLUMN_NAMES.get(name, name),
            ).in_(values)
        )
    return query


def _filter_query_for_user_flags(query, values, /, *, user_id):
    if values:
        if "NULL" in values:
            query = query.outerjoin(SqlUserFlag).filter(
                or_(
                    and_(
                        SqlUserFlag.user_id == user_id,
                        SqlUserFlag.type.in_(
                            list(
                                map(FLAGS_BY_NAME.get, values),
                            )
                        ),
                    ),
                    or_(SqlUserFlag.user_id != user_id, (SqlUserFlag.type.is_(null()))),
                )
            )
        else:
            query = query.join(SqlUserFlag).filter(
                SqlUserFlag.user_id == user_id,
                SqlUserFlag.type.in_(
                    list(
                        map(FLAGS_BY_NAME.get, values),
                    )
                ),
            )
    return query


def _filter_transactions_based_on_text(query, search_text: str, dialect_name: str):
    # Escape special characters in search_text
    search_text = re.sub(r"([-\*\(\)~\"@<>\^+]+)", r"", search_text)
    query = query.join(SqlMessage, isouter=True)
    # check dialect and use appropriate full text search
    if dialect_name == "mysql":
        return query.filter(
            text(
                f"MATCH ({SqlTransaction.__tablename__}.endpoint) "
                f"AGAINST (:search_text IN BOOLEAN MODE) OR "
                f"MATCH ({SqlMessage.__tablename__}.summary) "
                f"AGAINST (:search_text IN BOOLEAN MODE)",
            ).bindparams(bindparam("search_text", literal_column(f"'{search_text}*'")))
        )

    return query.filter(
        (SqlTransaction.endpoint.ilike(bindparam("search_text", f"%{search_text}%")))
        | SqlMessage.summary.ilike(bindparam("search_text", f"%{search_text}%"))
    )


class SqlStorage(IStorage):
    """
    Storage implementation using SQLAlchemy (async).

    Currently supported/tested database drivers:

    - asyncpg (postgresql+asyncpg://...)
    - aiomysql (mysql+aiomysql://...)
    - asyncmy (mysql+asyncmy://...)

    And also, for development/testing purposes:

    - aiosqlite (sqlite+aiosqlite://...)

    Other async postgresql drivers may work, but as they're not included in the test suite, you may have surprises.
    Contributions welcome.

    """

    engine: AsyncEngine
    """Reference to the sqlalchemy async engine, which is a gateway to the database connectivity, able to provide a
    connection used to execute queries."""

    def __init__(self, engine: AsyncEngine, blob_storage: IBlobStorage, settings: StorageSettings):
        # XXX lokks like settings are not used anymore, except to know if we should run migrations or not
        self._should_migrate = settings.migrate

        self.metadata = Base.metadata

        self.blob_storage = blob_storage

        self.engine = engine
        self.session_factory = async_sessionmaker(self.engine, expire_on_commit=False)

        self._is_ready = asyncio.Event()
        self._worker = None

        self.blobs = BlobsRepository(self.session_factory)
        self.messages = MessagesRepository(self.session_factory)
        self.tags = TagsRepository(self.session_factory)
        self.tag_values = TagValuesRepository(self.session_factory)
        self.transactions = TransactionsRepository(self.session_factory, tags=self.tags, tag_values=self.tag_values)
        self.users = UsersRepository(self.session_factory)
        self.metrics = MetricsRepository(self.session_factory)
        self.metric_values = MetricValuesRepository(self.session_factory)
        self.flags = FlagsRepository(self.session_factory)

        self._debug = False

    def __repr__(self):
        return f'<{type(self).__name__} dialect="{self.engine.dialect.name}" object at {hex(id(self))}>'

    @asynccontextmanager
    async def begin(self):
        async with self.session_factory() as session:
            async with session.begin():
                yield session

    def install_debugging_instrumentation(self, *, echo=False):
        self._debug = True
        self._original_session_factory = self.session_factory
        self._debug_index = 1
        self.sql_queries = []

        if echo:
            from rich.console import Console
            from rich.syntax import Syntax

            _console = Console(force_terminal=True, width=180)

        @asynccontextmanager
        async def _session_factory():
            async with self._original_session_factory() as session:
                original_execute = session.sync_session._execute_internal

                def _instrumented_execute(statement, *args, **kwargs):
                    sql_query = str(
                        statement.compile(
                            session.sync_session.bind,
                            compile_kwargs={"literal_binds": True},
                        )
                    )
                    self.sql_queries.append(sql_query)

                    final_result = original_execute(statement, *args, **kwargs)

                    if echo:
                        _console.print(f"🛢 SQL QUERY (#{self._debug_index})", style="bold")
                        _console.print(
                            Syntax(
                                sql_query,
                                "sql",
                                word_wrap=True,
                                theme="vs",
                            )
                        )

                    if echo and self.engine.dialect.name == "postgresql":
                        _get0 = itemgetter(0)
                        _console.print(f"🛢 EXPLAIN ANALYZE (#{self._debug_index})", style="bold")
                        _console.print(
                            Syntax(
                                "\n".join(
                                    map(
                                        _get0,
                                        original_execute(text("EXPLAIN ANALYZE " + sql_query)).fetchall(),
                                    )
                                ),
                                "sql",
                                word_wrap=True,
                                theme="vs",
                            )
                        )

                    self._debug_index += 1
                    return final_result

                session.sync_session._execute_internal = _instrumented_execute
                try:
                    yield session
                finally:
                    session.sync_session._execute_internal = original_execute

        self.session_factory = _session_factory

    async def initialize(self):
        """Initialize database."""
        await self.create_users(["anonymous"])
        self._is_ready.set()

    async def finalize(self):
        await self.worker.close()
        await self.engine.dispose()

    @property
    def worker(self):
        if not self._worker:
            self._worker = AsyncWorkerQueue()
        return self._worker

    @override
    async def get_facet_meta(self, name):
        if name == "endpoint":
            # get transaction count grouped by endpoint
            async with self.begin() as session:
                query = select(SqlTransaction.endpoint, func.count()).group_by(SqlTransaction.endpoint)
                result = await session.execute(query)
                return {row[0]: row[1] for row in result.fetchall()}

        raise NotImplementedError(f"Unknown facet: {name}")

    @override
    async def get_transaction_list(
        self,
        *,
        username: str,
        with_messages=False,
        filters=None,
        page: int = 1,
        cursor: str = "",
        text_search="",
    ):
        """
        Implements :meth:`Storage.find_transactions <harp.typing.storage.Storage.find_transactions>`.

        """

        user = await self.users.find_one_by_username(username)

        result = Results()
        query = self.transactions.select(
            with_messages=with_messages,
            with_user_flags=user.id if user else False,
            with_tags=True,
        )

        if filters:
            query = _filter_query(query, "endpoint", filters.get("endpoint", None))
            query = _filter_query(query, "method", filters.get("method", None))
            query = _filter_query(query, "status", filters.get("status", None))
            query = _filter_query_for_user_flags(query, filters.get("flag", None), user_id=user.id)
            query = _numerical_filter_query(query, "tpdex", filters.get("tpdex", None))

        if text_search:
            query = _filter_transactions_based_on_text(query, text_search, dialect_name=self.engine.dialect.name)

        query = query.order_by(SqlTransaction.started_at.desc())

        # apply cursor (before count)
        if page and cursor:
            query = query.filter(SqlTransaction.id <= cursor)

        async with self.begin() as session:
            # count items from query
            result.meta["total"] = await session.scalar(
                query.with_only_columns(func.count(SqlTransaction.id)).order_by(None)
            )

        # apply limit/offset (after count)
        query = query.limit(PAGE_SIZE)
        if page:
            query = query.offset(max(0, (page - 1) * PAGE_SIZE))

        async with self.begin() as session:
            for transaction in (await session.scalars(query)).unique().all():
                result.append(transaction.to_model(with_user_flags=True))

        return result

    @override
    async def get_transaction(self, id: str, /, *, username: str) -> Optional[Transaction]:
        user = await self.users.find_one_by_username(username)

        return (
            await self.transactions.find_one_by_id(
                id,
                with_messages=True,
                with_user_flags=user.id if user else False,
                with_tags=True,
            )
        ).to_model(with_user_flags=True)

    @override
    async def transactions_grouped_by_time_bucket(
        self,
        *,
        endpoint: Optional[str] = None,
        time_bucket: str = TimeBucket.DAY.value,
        start_datetime: Optional[datetime] = None,
    ) -> list[TransactionsGroupedByTimeBucket]:
        if time_bucket not in [e.value for e in TimeBucket]:
            raise ValueError(
                f"Invalid time bucket: {time_bucket}. Must be one of {', '.join([e.value for e in TimeBucket])}."
            )

        c_started_at = SqlTransaction.started_at
        s_date = TruncDatetime(literal(time_bucket), c_started_at).label("tb")
        query = select(
            s_date,
            func.count(),
            func.sum(case((SqlTransaction.x_status_class.in_(("5xx", "ERR")), 1), else_=0)),
            func.sum(
                case(
                    (
                        and_(
                            SqlTransaction.x_cached is not None,
                            SqlTransaction.x_cached != "",
                        ),
                        1,
                    ),
                    else_=0,
                )
            ),
            func.avg(SqlTransaction.elapsed),
            func.avg(SqlTransaction.tpdex),
        )

        if endpoint:
            query = query.where(SqlTransaction.endpoint == endpoint)

        if start_datetime:
            query = query.where(c_started_at >= start_datetime.astimezone(UTC))

        query = query.group_by(s_date).order_by(s_date.asc())
        async with self.begin() as session:
            result = await session.execute(query)
            return [
                TransactionsGroupedByTimeBucket(
                    datetime=ensure_datetime(row[0], UTC),
                    count=row[1],
                    errors=int(row[2]),
                    cached=int(row[3]),
                    meanDuration=row[4] if row[4] else 0,
                    meanTpdex=row[5],
                )
                for row in result.fetchall()
            ]

    async def get_usage(self):
        async with self.begin() as session:
            query = select(count(SqlTransaction.id)).where(
                SqlTransaction.started_at > (datetime.now(UTC) - timedelta(hours=24))
            )
            return (await session.execute(query)).scalar_one_or_none()

    async def _on_startup_actions(self, TransactionEvent):
        """Event handler to create the database tables on startup. May drop them first if configured to do so."""
        await self.initialize()

    @override
    async def set_user_flag(self, *, transaction_id: str, username: str, flag: int, value=True):
        """Sets or unsets a user flag on a transaction."""
        async with self.begin() as session:
            user = await self.users.find_one_by_username(username)
            transaction = await self.transactions.find_one_by_id(transaction_id)

            if value:
                session.add(
                    SqlUserFlag(
                        transaction_id=transaction.id,
                        user_id=user.id,
                        type=flag,
                    )
                )
            else:
                await session.execute(
                    delete(SqlUserFlag).where(
                        SqlUserFlag.transaction_id == transaction.id,
                        SqlUserFlag.user_id == user.id,
                    )
                )

    async def ready(self):
        await self._is_ready.wait()

    @override
    async def create_users_once_ready(self, users: Iterable[str]):
        """Sets the list of users to be created once the database is ready."""

        async def defered_create_users():
            await self.ready()
            await self.create_users(users)

        await self.worker.push(defered_create_users)

    async def create_users(self, users: Iterable[str]):
        async with self.begin() as session:
            for username in users:
                # Check if the username already exists
                result = await session.execute(select(SqlUser).where(SqlUser.username == username))
                existing_user = result.scalar_one_or_none()

                # If the username does not exist, create a new user
                if existing_user is None:
                    user = SqlUser()
                    user.username = username
                    session.add(user)



from typing import cast
from unittest.mock import ANY, AsyncMock

import orjson
import pytest
import respx
from httpx import AsyncClient, Response
from whistle import AsyncEventDispatcher

from harp.asgi.events import EVENT_CORE_STARTED
from harp.typing import GlobalSettings
from harp_apps.telemetry.manager import TelemetryManager

stub_settings = cast(GlobalSettings, {"applications": ["oh", "my"]})


class MockedTelemetryManager(TelemetryManager):
    start = AsyncMock()


class TestTelemetryManager:
    async def test_events(self):
        dispatcher = AsyncEventDispatcher()
        assert not dispatcher.has_listeners(EVENT_CORE_STARTED)
        manager = MockedTelemetryManager(stub_settings, AsyncClient(), None, dispatcher)
        assert dispatcher.has_listeners(EVENT_CORE_STARTED)

        assert not manager.start.called
        await dispatcher.adispatch(EVENT_CORE_STARTED)
        assert manager.start.called and manager.start.call_count == 1

    @respx.mock
    @pytest.mark.parametrize("url", [None, "https://example.com/"])
    async def test_ping(self, url):
        endpoint = respx.post(url).mock(return_value=Response(201, content=b"{}"))

        manager = TelemetryManager(stub_settings, AsyncClient(), endpoint=url)

        assert not endpoint.called

        await manager.ping()

        assert endpoint.called and endpoint.call_count == 1
        assert endpoint.calls.last.request.url == (url or manager.default_endpoint)
        assert endpoint.calls.last.request.method == "POST"
        assert orjson.loads(endpoint.calls.last.request.content) == {
            "f": ANY,
            "a": "oh,my",
            "v": ANY,
            "c": 0,
            "i": 0,
            "t": "start",
        }
        assert manager.count == 1



from functools import reduce

import orjson

from harp_apps.rules.constants import DEFAULT_LEVELS, DEFAULT_RULES_LEVELS
from harp_apps.rules.models.compilers import BaseRuleSetCompiler


def _match_level(rules, against):
    for pattern, rules in rules:
        if pattern.match(against):
            if hasattr(rules, "items"):
                yield from rules.items()
            else:
                yield from rules


def _rules_as_human_dict(rules: dict, *, show_scripts=True):
    result = {}
    for k, v in rules.items():
        if isinstance(v, dict):
            result[k.source] = _rules_as_human_dict(v, show_scripts=show_scripts)
        else:
            if show_scripts:
                result[k.source] = [script.source.strip() for script in v]
                if len(result[k.source]) == 1:
                    result[k.source] = result[k.source][0]
            else:
                result[k.source] = "..."
    return result


def _recursive_len(rules: dict):
    return reduce(
        lambda x, y: x + _recursive_len(y[1]) if hasattr(y[1], "items") else x + 1,
        rules.items(),
        0,
    )


class BaseRuleSet:
    CompilerType = BaseRuleSetCompiler

    def __init__(self, rules=None, /, *, levels=DEFAULT_LEVELS):
        #: levels: tuple of the rules hierarchy level names
        self._levels = levels

        #: compiler: compiler implementation
        self._compiler = self.CompilerType(levels=levels)

        #: rules: the compiled rules
        self._rules = rules or {}

    @property
    def rules(self):
        return self._rules

    def add(self, sources: dict):
        self._rules = self._compiler.compile(sources, target=self._rules)

    def match(self, *args):
        """
        Match the given arguments against the rules. Each argument must match a "level" in this ruleset.
        """
        if len(args) != len(self._levels):
            raise ValueError(f"Expected {len(self._levels)} arguments, got {len(args)}")

        rules = self.rules.items()
        for against in args:
            rules = _match_level(rules, against)
        yield from rules

    def _asdict(self, /, *, secure=True):
        return _rules_as_human_dict(self.rules)

    def __len__(self):
        return _recursive_len(self.rules)

    def __repr__(self):
        return f"{type(self).__name__}({orjson.dumps(_rules_as_human_dict(self.rules, show_scripts=False)).decode()})"


class RuleSet(BaseRuleSet):
    def __init__(self):
        super().__init__(levels=DEFAULT_RULES_LEVELS)



import time
from functools import wraps
from typing import Optional

import httpx

from harp.controllers.typing import IAsyncController
from harp.http import HttpRequest, HttpResponse
from harp_apps.metrics.constants import (
    CONTROLLER_EXCEPTIONS,
    CONTROLLER_REQUESTS,
    CONTROLLER_REQUESTS_IN_PROGRESS,
    CONTROLLER_REQUESTS_TIME,
    CONTROLLER_RESPONSES,
    REMOTE_EXCEPTIONS,
    REMOTE_REQUESTS,
    REMOTE_REQUESTS_IN_PROGRESS,
    REMOTE_REQUESTS_TIME,
    REMOTE_RESPONSES,
)


def decorate_proxy_controller(wrapped: IAsyncController, *, name: Optional[str] = None, port: Optional[int] = None):
    """
    Decorates a builtin or custom proxy controller to collect metrics for prometheus.
    """
    common_labels = dict(name=(name or "-"), port=(port or "-"))

    @wraps(wrapped)
    async def wrapper(request: HttpRequest, *args, **kwargs) -> HttpResponse:
        labels = {**common_labels, "method": request.method}
        CONTROLLER_REQUESTS.labels(**labels).inc()
        CONTROLLER_REQUESTS_IN_PROGRESS.labels(**labels).inc()

        before_time = time.perf_counter()
        try:
            response = await wrapped(request, *args, **kwargs)
            CONTROLLER_RESPONSES.labels(**labels, status=getattr(response, "status", "-")).inc()
            return response
        except Exception as exc:  # noqa: BLE001
            CONTROLLER_EXCEPTIONS.labels(**labels, path=request.path, exception=type(exc).__name__).inc()
            raise exc from None
        finally:
            after_time = time.perf_counter()
            CONTROLLER_REQUESTS_TIME.labels(**labels).observe(after_time - before_time)
            CONTROLLER_REQUESTS_IN_PROGRESS.labels(**labels).dec()

    return wrapper


def decorate_http_client_send(wrapped):
    """
    Decorates the httpx async client send method to collect metrics for prometheus.
    """

    @wraps(wrapped)
    async def wrapper(request: httpx.Request, *args, **kwargs) -> httpx.Response:
        labels = {"name": request.extensions.get("harp", {}).get("endpoint", "-"), "method": request.method}
        REMOTE_REQUESTS.labels(**labels).inc()
        REMOTE_REQUESTS_IN_PROGRESS.labels(**labels).inc()

        before_time = time.perf_counter()
        try:
            response = await wrapped(request, *args, **kwargs)
            REMOTE_RESPONSES.labels(**labels, status=getattr(response, "status_code", "-")).inc()
            return response
        except Exception as exc:  # noqa: BLE001
            REMOTE_EXCEPTIONS.labels(**labels, url=str(request.url), exception=type(exc).__name__).inc()
            raise exc from None
        finally:
            after_time = time.perf_counter()
            REMOTE_REQUESTS_TIME.labels(**labels).observe(after_time - before_time)
            REMOTE_REQUESTS_IN_PROGRESS.labels(**labels).dec()

    return wrapper



from typing import Optional

from .blobs import BlobStorageSettings
from .database import DatabaseSettings
from .redis import RedisSettings


class StorageSettings(DatabaseSettings):
    migrate: bool = True
    blobs: BlobStorageSettings = BlobStorageSettings()
    redis: Optional[RedisSettings] = None
